from datetime import datetime, timedelta, timezone
import json
import os

import pytest

from livestyled.models.booking import Booking
from livestyled.models.device import Device
from livestyled.models.event import Event
from livestyled.models.user import User
from livestyled.resource_client import LiveStyledResourceClient
from livestyled.tests.utils import configure_mock_responses

FIXTURES_DIR = os.path.join(os.path.dirname(__file__), 'fixtures')
TEST_API_DOMAIN = 'test.livestyled.com'
CONTENT_TYPE = 'application/ld+json'


def test_get_bookings(requests_mock):
    mock_responses = (
        ('GET', 'https://' + TEST_API_DOMAIN + '/v4/bookings', 'mock_responses/ls_api/bookings_page1.json', 200),
        ('GET', 'https://' + TEST_API_DOMAIN + '/v4/bookings?page=2', 'mock_responses/ls_api/bookings_page2.json', 200),
    )
    configure_mock_responses(requests_mock, mock_responses, FIXTURES_DIR, CONTENT_TYPE)

    resource_client = LiveStyledResourceClient(TEST_API_DOMAIN, 'bar')
    bookings = list(resource_client.get_bookings())
    assert bookings
    assert len(bookings) == 60
    booking = bookings[0]
    assert isinstance(booking, Booking)
    assert booking.type == 'e-ticket'
    assert booking.id == 16
    assert booking.device
    assert isinstance(booking.device, Device)
    assert booking.device.id == 1748495
    assert booking.user
    assert isinstance(booking.user, User)
    assert booking.user.id == 276270
    assert booking.event
    assert isinstance(booking.event, Event)
    assert booking.event.id == 100015938
    assert booking.created_at == datetime(2019, 8, 16, 9, 58, 35, tzinfo=timezone(timedelta(0), '+0000'))
    assert booking.updated_at == datetime(2019, 8, 16, 9, 58, 35, tzinfo=timezone(timedelta(0), '+0000'))
    assert booking.action == 'going'

    booking = bookings[3]
    assert isinstance(booking, Booking)
    assert booking.type == 'manual'
    assert booking.id == 53
    assert booking.device
    assert isinstance(booking.device, Device)
    assert booking.device.id == 1748495
    assert not booking.user
    assert booking.event
    assert isinstance(booking.event, Event)
    assert booking.event.id == 100015860
    assert booking.created_at == datetime(2019, 8, 21, 13, 31, 47, tzinfo=timezone(timedelta(0), '+0000'))
    assert booking.updated_at == datetime(2019, 8, 21, 13, 31, 47, tzinfo=timezone(timedelta(0), '+0000'))
    assert booking.action == 'interested'


@pytest.mark.parametrize(('device', 'user', 'event'), [
    ('1234', '5555', '8888'),
    (Device.placeholder(1234), User.placeholder(5555), Event.placeholder(8888))
])
def test_create_booking(requests_mock, device, user, event):

    mock_responses = (
        ('POST', 'https://' + TEST_API_DOMAIN + '/v4/bookings', 'mock_responses/ls_api/create_booking.json', 200),
    )
    configure_mock_responses(requests_mock, mock_responses, FIXTURES_DIR, CONTENT_TYPE)

    resource_client = LiveStyledResourceClient(TEST_API_DOMAIN, 'bar')
    resource_client.create_booking(
        Booking.create_new(
            type='manual',
            device=device,
            user=user,
            event=event,
            action='going'
        )
    )

    assert len(requests_mock.request_history) == 1
    assert requests_mock.request_history[0].method == 'POST'
    assert requests_mock.request_history[0].url == 'https://' + TEST_API_DOMAIN + '/v4/bookings'
    assert json.loads(requests_mock.request_history[0].body) == {
        'action': 'going',
        'event': '/v4/events/8888',
        'device': '/v4/devices/1234',
        'type': 'manual',
        'user': '/v4/users/5555'
    }



from livestyled.models.app import App, Currency
from livestyled.models.booking import Booking
from livestyled.models.cohort import Cohort
from livestyled.models.competition import Competition
from livestyled.models.device import Device
from livestyled.models.device_consent import DeviceConsent
from livestyled.models.device_preference import DevicePreference
from livestyled.models.device_push_consent import DevicePushConsent
from livestyled.models.device_token import DeviceToken
from livestyled.models.event import Event
from livestyled.models.fixture import Fixture
from livestyled.models.fulfilment_point import (
    FulfilmentPoint,
    FulfilmentPointCategory,
    FulfilmentPointCategoryTranslation,
    FulfilmentPointTranslation
)
from livestyled.models.league_table import LeagueTable, LeagueTableGroup
from livestyled.models.magic_field import MagicField
from livestyled.models.news import News
from livestyled.models.order import Order, OrderItem
from livestyled.models.product import (
    Product,
    ProductCategory,
    ProductImage,
    ProductTranslation,
    ProductVariant
)
from livestyled.models.push_broadcast import PushBroadcast
from livestyled.models.push_consent import PushConsent
from livestyled.models.season import Season
from livestyled.models.sport_venue import SportVenue
from livestyled.models.team import Team
from livestyled.models.ticket import Ticket
from livestyled.models.ticket_auth import TicketAuth
from livestyled.models.ticket_integration import TicketIntegration
from livestyled.models.user import User, UserEmail, UserInfo, UserSSO
from livestyled.models.venue import Venue

__all__ = [
    App,
    Currency,
    Booking,
    Cohort,
    Competition,
    Device,
    DeviceConsent,
    DevicePreference,
    DevicePushConsent,
    DeviceToken,
    Event,
    Fixture,
    FulfilmentPoint,
    FulfilmentPointCategory,
    FulfilmentPointCategoryTranslation,
    FulfilmentPointTranslation,
    LeagueTable,
    LeagueTableGroup,
    MagicField,
    News,
    Order,
    OrderItem,
    Product,
    ProductCategory,
    ProductImage,
    ProductTranslation,
    ProductVariant,
    PushBroadcast,
    PushConsent,
    Season,
    SportVenue,
    Team,
    Ticket,
    TicketAuth,
    TicketIntegration,
    UserEmail,
    UserInfo,
    User,
    UserSSO,
    Venue,
]



from marshmallow import EXCLUDE, fields, Schema

from livestyled.models import Event
from livestyled.schemas.fields import RelatedResourceLinkField
from livestyled.schemas.utils import get_id_from_url
from livestyled.schemas.venue import VenueSchema


class EventDate(Schema):
    class Meta:
        unknown = EXCLUDE

    id = fields.Int()
    start_at = fields.AwareDateTime(data_key='startAt')
    end_at = fields.AwareDateTime(data_key='endAt')
    general_ticket_url = fields.String(data_key='generalTicketUrl', required=False)


class CoreEventCategory(Schema):
    class Meta:
        unknown = EXCLUDE

    name = fields.String()


class EventSchema(Schema):
    class Meta:
        unknown = EXCLUDE
        api_type = 'events'
        url = 'v4/events'
        model = Event

    id = fields.Int()
    status = fields.String()
    type = fields.String()
    title = fields.String()
    description = fields.String()
    image_url = fields.Url(data_key='imageUrl')
    promoted = fields.Boolean()
    event_dates = fields.Nested(EventDate, data_key='eventDates', many=True)
    app = fields.Function(get_id_from_url)
    venues = fields.List(RelatedResourceLinkField(schema=VenueSchema), many=True, missing=None)
    artists = fields.List(fields.Inferred),  # TODO
    useful_info = fields.List(fields.Inferred, data_key='usefulInfo'),  # TODO
    user_actions = fields.List(fields.Inferred, data_key='userActions'),  # TODO
    social_media = fields.List(fields.Inferred, data_key='socialMedia'),  # TODO
    translations = fields.List(fields.Inferred),  # TODO
    updated_at = fields.AwareDateTime(data_key='updatedAt', allow_none=True)
    created_at = fields.AwareDateTime(data_key='createdAt', allow_none=True)
    core_event_category = fields.Nested(CoreEventCategory, data_key='coreEventCategory', missing=None)



from livestyled.schemas.app import AppSchema
from livestyled.schemas.booking import BookingSchema
from livestyled.schemas.cohort import CohortSchema
from livestyled.schemas.competition import CompetitionSchema
from livestyled.schemas.device import DeviceSchema
from livestyled.schemas.device_consent import DeviceConsentSchema
from livestyled.schemas.device_preference import DevicePreferenceSchema
from livestyled.schemas.device_token import DeviceTokenSchema
from livestyled.schemas.event import EventSchema
from livestyled.schemas.event_category import EventCategorySchema
from livestyled.schemas.event_integration import EventIntegrationSchema
from livestyled.schemas.event_stage import EventStageSchema
from livestyled.schemas.fixture import FixtureSchema
from livestyled.schemas.league_table import LeagueTableGroupSchema, LeagueTableSchema
from livestyled.schemas.magic_field import MagicFieldSchema
from livestyled.schemas.news import NewsSchema
from livestyled.schemas.order import OrderSchema
from livestyled.schemas.product import ProductCategorySchema, ProductSchema
from livestyled.schemas.push_broadcast import PushBroadcastSchema
from livestyled.schemas.push_consent import PushConsentSchema
from livestyled.schemas.season import SeasonSchema
from livestyled.schemas.sport_venue import SportVenueSchema
from livestyled.schemas.team import TeamSchema
from livestyled.schemas.ticket import TicketSchema
from livestyled.schemas.ticket_auth import TicketAuthSchema
from livestyled.schemas.ticket_integration import TicketIntegrationSchema
from livestyled.schemas.user import UserCreateSchema, UserEmailSchema, UserInfoSchema, UserSchema, UserSSOSchema
from livestyled.schemas.venue import VenueSchema

__all__ = [
    AppSchema,
    BookingSchema,
    CohortSchema,
    CompetitionSchema,
    DeviceSchema,
    DeviceConsentSchema,
    DevicePreferenceSchema,
    DeviceTokenSchema,
    EventSchema,
    EventCategorySchema,
    EventIntegrationSchema,
    EventStageSchema,
    FixtureSchema,
    LeagueTableSchema,
    LeagueTableGroupSchema,
    MagicFieldSchema,
    NewsSchema,
    OrderSchema,
    ProductCategorySchema,
    ProductSchema,
    PushBroadcastSchema,
    PushConsentSchema,
    SeasonSchema,
    SportVenueSchema,
    TeamSchema,
    TicketAuthSchema,
    TicketIntegrationSchema,
    TicketSchema,
    UserCreateSchema,
    UserEmailSchema,
    UserInfoSchema,
    UserSchema,
    UserSSOSchema,
    VenueSchema,
]



from datetime import datetime


class Venue:
    def __init__(
            self,
            id: int,
            name: str or None,
            label: str or None,
            status: str or None,
            is_default: bool,
            description: str or None = None,
            image_url: str or None = None,
            map_image_url: str or None = None,
            geo_latitude: str or None = None,
            geo_longitude: str or None = None,
            geo_latitude_north_west: str or None = None,
            geo_longitude_north_west: str or None = None,
            geo_latitude_south_east: str or None = None,
            geo_longitude_south_west: str or None = None,
            city: str or None = None,
            external_id: str or None = None,
            created_at: datetime or None = None,
            updated_at: datetime or None = None,
            venue_icon_url: str or None = None,
    ):
        self.id = id
        self.name = name
        self.label = label
        self.status = status
        self.is_default = is_default
        self.description = description
        self.image_url = image_url
        self.map_image_url = map_image_url
        self.geo_latitude = geo_latitude
        self.geo_longitude = geo_longitude
        self.geo_latitude_north_west = geo_latitude_north_west
        self.geo_longitude_north_west = geo_longitude_north_west
        self.geo_latitude_south_east = geo_latitude_south_east
        self.geo_longitude_south_west = geo_longitude_south_west
        self.city = city
        self.external_id = external_id
        self.created_at = created_at
        self.updated_at = updated_at
        self.venue_icon_url = venue_icon_url

    @classmethod
    def create_new(
            cls,
            name: str,
            label: str or None,
            status: str or None,
            is_default: bool,
            description: str or None = None,
            image_url: str or None = None,
            map_image_url: str or None = None,
            geo_latitude: str or None = None,
            geo_longitude: str or None = None,
            geo_latitude_north_west: str or None = None,
            geo_longitude_north_west: str or None = None,
            geo_latitude_south_east: str or None = None,
            geo_longitude_south_west: str or None = None,
            city: str or None = None,
            external_id: str or None = None,
            created_at: datetime or None = None,
            updated_at: datetime or None = None,
            venue_icon_url: str or None = None,
    ):
        return Venue(
            id=None,
            name=name,
            label=label,
            status=status,
            is_default=is_default,
            description=description,
            image_url=image_url,
            map_image_url=map_image_url,
            geo_latitude=geo_latitude,
            geo_longitude=geo_longitude,
            geo_latitude_north_west=geo_latitude_north_west,
            geo_longitude_north_west=geo_longitude_north_west,
            geo_latitude_south_east=geo_latitude_south_east,
            geo_longitude_south_west=geo_longitude_south_west,
            city=city,
            external_id=external_id,
            created_at=created_at,
            updated_at=updated_at,
            venue_icon_url=venue_icon_url,
        )

    @classmethod
    def placeholder(
            cls,
            id
    ):
        return cls(
            id=id,
            name=None,
            label=None,
            status=None,
            is_default=False
        )



from datetime import datetime, timedelta, timezone

from livestyled.models import Device, DeviceConsent
from livestyled.utils import create_resource_from_data


def test_create_resource_from_data():
    data = {
        '@context': '/v4/contexts/Device',
        '@id': '/v4/devices/2659',
        '@type': 'Device',
        'id': 2659,
        'token': 'EE76F4D8-BC5A-438D-9F06-9F6873F60960',
        'consent': {
            'id': 3248,
            'locationCapture': False,
            'locationGranular': None,
            'camera': False,
            'calendar': False,
            'photoSharing': False,
            'pushNotification': True,
            'createdAt': '2019-10-07T15:43:18+00:00',
            'updatedAt': '2019-10-07T15:43:18+00:00'
        },
        'type': 'IOS',
        'status': 'ACTIVE',
        'appVersion': '8.1',
        'osVersion': None,
        'model': 'iPhone 5 (GSM+CDMA)',
        'manufacturer': 'Apple',
        'bluetoothOn': None,
        'wifiConnected': None,
        'createdAt': None,
        'updatedAt': None
    }
    device = create_resource_from_data(Device, data)
    assert isinstance(device, Device)
    assert device.id == 2659
    assert isinstance(device.consent, DeviceConsent)
    assert device.consent.created_at == datetime(2019, 10, 7, 15, 43, 18, tzinfo=timezone(timedelta(0), '+0000'))



class PushConsent:
    def __init__(
            self,
            id: int or None,
            type: str or None,
            title: str or None,
            label: str or None,
            sort_id: int or None,
            updated_at,
            created_at,
            translations
    ):
        self.id = id
        self.type = type
        self.title = title
        self.label = label
        self.sort_id = sort_id
        self.updated_at = updated_at
        self.created_at = created_at
        self.translations = translations

    @classmethod
    def create_new(
            cls,
            type,
            title,
            label,
            sort_id,
            updated_at,
            created_at,
            translations
    ):
        return PushConsent(
            id=None,
            type=type,
            label=label,
            title=title,
            sort_id=sort_id,
            updated_at=updated_at,
            created_at=created_at,
            translations=translations
        )

    @classmethod
    def placeholder(
            cls,
            id
    ):
        return cls(
            id=id,
            type=None,
            label=None,
            title=None,
            sort_id=None,
            updated_at=None,
            created_at=None,
            translations=None
        )

    def diff(self, other):
        differences = {}
        fields = (
            'type', 'title', 'sort_id', 'updated_at', 'created_at', 'translations', 'label'
        )
        for field in fields:
            if getattr(self, field) != getattr(other, field):
                differences[field] = getattr(self, field)
        return differences



import datetime
import os

from livestyled.models.cohort import Cohort
from livestyled.models.user import User
from livestyled.resource_client import LiveStyledResourceClient
from livestyled.schemas.user import UserSchema
from livestyled.tests.utils import configure_mock_responses

FIXTURES_DIR = os.path.join(os.path.dirname(__file__), 'fixtures')
TEST_API_DOMAIN = 'test.livestyled.com'
CONTENT_TYPE = 'application/ld+json'


def test_get_user_by_id(requests_mock):
    mock_responses = (
        ('GET', 'https://' + TEST_API_DOMAIN + '/v4/users/1234', 'mock_responses/ls_api/user_1234.json', 200),
    )
    configure_mock_responses(requests_mock, mock_responses, FIXTURES_DIR, CONTENT_TYPE)

    resource_client = LiveStyledResourceClient(TEST_API_DOMAIN, 'bar')
    user = resource_client.get_user(1234)
    assert user
    assert isinstance(user, User)
    assert user.id == 1234
    assert user.email == 'bob.bobberson@test.com'
    assert user.auth_type == 'LOCAL'
    assert user.first_name == 'Bob'
    assert user.last_name == 'Bobberson'
    assert user.password is None
    assert set(user.cohorts) == {
        Cohort.placeholder(1),
        Cohort.placeholder(2),
        Cohort.placeholder(8)
    }
    assert user.user_info.first_name == 'Bob'
    assert user.user_info.last_name == 'Bobberson'
    assert user.user_info.dob == datetime.datetime(1992, 12, 6, 0, 0, tzinfo=datetime.timezone(datetime.timedelta(0), '+0000'))
    assert user.user_info.gender == 'MALE'
    assert len(user.magic_fields) == 1
    assert user.magic_fields[0].key == 'rtest'
    assert user.magic_fields[0].value == 'foobar'
    assert user.magic_fields[0].id == 52
    assert len(user.devices) == 1
    assert user.devices[0].id == 1753940


def test_user_diff(requests_mock):
    mock_responses = (
        ('GET', 'https://' + TEST_API_DOMAIN + '/v4/users/1234', 'mock_responses/ls_api/user_1234.json', 200),
    )
    configure_mock_responses(requests_mock, mock_responses, FIXTURES_DIR, CONTENT_TYPE)

    resource_client = LiveStyledResourceClient(TEST_API_DOMAIN, 'bar')
    user1 = resource_client.get_user(1234)
    user2 = resource_client.get_user(1234)

    user2.user_info.first_name = 'Bobby'

    differences = user2.diff(user1)
    assert differences
    assert 'user_info' in differences

    serialised_differences = UserSchema(only=['user_info']).dump(differences)
    assert serialised_differences
    assert serialised_differences == {
        'userInfo': {
            'id': 37,
            'dob': '1992-12-06T00:00:00+00:00',
            'firstName': 'Bobby',
            'gender': 'MALE',
            'lastName': 'Bobberson',
            'phone': None
        }
    }



from livestyled.models.competition import Competition
from livestyled.models.season import Season
from livestyled.models.team import Team


class LeagueTableGroup:
    def __init__(
            self,
            id,
            reference,
            title,
    ):
        self._id = id
        self._reference = reference
        self._title = title

    @classmethod
    def create_new(
            cls,
            reference,
            title
    ):
        return LeagueTableGroup(
            id=None,
            reference=reference,
            title=title
        )

    @property
    def id(self):
        return self._id

    @property
    def reference(self):
        return self._reference

    @property
    def title(self):
        return self._title

    @classmethod
    def placeholder(
            cls,
            id
    ):
        return cls(
            id=id,
            reference=None,
            title=None
        )

    def __repr__(self):
        return '<LeagueTableGroup(id={self.id!r})>'.format(self=self)

    def diff(self, other):
        differences = {}
        fields = (
            'reference', 'title'
        )
        for field in fields:
            if getattr(self, field) != getattr(other, field):
                differences[field] = getattr(self, field)
        return differences


class LeagueTable:
    def __init__(
            self,
            id,
            external_id,
            team_id,
            featured_team,
            position,
            played,
            goals_for,
            goals_against,
            start_day_position,
            won,
            lost,
            drawn,
            goal_difference,
            points,
            competition_id,
            season_id,
            group_id=None
    ):
        self._id = id
        self.external_id = external_id
        self._team = Team.placeholder(id=team_id)
        self._season = Season.placeholder(id=season_id)
        self._competition = Competition.placeholder(id=competition_id)
        if group_id:
            self._group = LeagueTableGroup.placeholder(id=group_id)
        else:
            self._group = None
        self.featured_team = featured_team
        self.position = position
        self.played = played
        self.goals_for = goals_for
        self.goals_against = goals_against
        self.start_day_position = start_day_position
        self.won = won
        self.lost = lost
        self.drawn = drawn
        self.goal_difference = goal_difference
        self.points = points

    @classmethod
    def create_new(
            cls,
            external_id,
            team: Team,
            featured_team,
            position,
            played,
            goals_for,
            goals_against,
            start_day_position,
            won,
            lost,
            drawn,
            goal_difference,
            points,
            competition: Competition,
            season: Season,
            group
    ):
        league_table = LeagueTable(
            id=None,
            external_id=external_id,
            featured_team=featured_team,
            position=position,
            played=played,
            goals_for=goals_for,
            goals_against=goals_against,
            start_day_position=start_day_position,
            won=won,
            lost=lost,
            drawn=drawn,
            goal_difference=goal_difference,
            points=points,
            season_id=None,
            competition_id=None,
            group_id=None,
            team_id=None
        )
        league_table._team = team
        league_table._season = season
        league_table._competition = competition
        league_table._group = group
        return league_table

    @property
    def id(self):
        return self._id

    @property
    def competition_id(self):
        return self._competition.id

    @property
    def team_id(self):
        return self._team.id

    @property
    def season_id(self):
        return self._season.id

    @property
    def group_id(self):
        if self._group:
            return self._group.id
        else:
            return None

    def __repr__(self):
        return '<LeagueTable(id={self.id!r})>'.format(self=self)

    def diff(self, other):
        differences = {}
        fields = (
            'competition_id', 'team_id', 'group_id', 'season_id',
            'featured_team', 'position', 'played', 'goals_for', 'goals_against', 'external_id',
            'start_day_position', 'won', 'lost', 'drawn', 'goal_difference', 'points'
        )
        for field in fields:
            if getattr(self, field) != getattr(other, field):
                differences[field] = getattr(self, field)
        return differences



from datetime import datetime, timedelta, timezone
import os

from livestyled.models import App, Currency, FulfilmentPoint, FulfilmentPointCategory, FulfilmentPointCategoryTranslation, FulfilmentPointTranslation, Order, OrderItem, Product, ProductCategory, ProductVariant, User
from livestyled.resource_client import LiveStyledResourceClient
from livestyled.tests.utils import configure_mock_responses

FIXTURES_DIR = os.path.join(os.path.dirname(__file__), 'fixtures')
TEST_API_DOMAIN = 'test.livestyled.com'
CONTENT_TYPE = 'application/ld+json'


def test_get_orders(requests_mock):
    mock_responses = (
        ('GET', 'https://' + TEST_API_DOMAIN + '/v4/orders', 'mock_responses/ls_api/orders.json', 200),
    )
    configure_mock_responses(requests_mock, mock_responses, FIXTURES_DIR, CONTENT_TYPE)

    resource_client = LiveStyledResourceClient(TEST_API_DOMAIN, 'bar')
    orders = list(resource_client.get_orders())
    assert len(orders) == 30
    orders.sort(key=lambda o: o.id)
    order = orders[0]
    assert isinstance(order, Order)
    assert order.id == 1
    assert order.status == 'COMPLETE'
    assert order.gross_amount == 1
    assert order.discount == 0
    assert order.net_amount == 1
    assert order.order_number == 12345678
    assert order.created_at == datetime(2019, 7, 4, 7, 43, 53, tzinfo=timezone(timedelta(0), '+0000'))
    assert order.updated_at == datetime(2020, 5, 27, 8, 48, 14, tzinfo=timezone(timedelta(0), '+0000'))
    assert len(order.items) == 1
    assert isinstance(order.items[0], OrderItem)
    order_item = order.items[0]
    assert order_item.id == 1
    assert order_item.quantity == 1
    assert order_item.title == 'Kartiks 1st Product'
    assert order_item.subtitle == 'TEST Product'
    assert order_item.image_url == 'https://apiv3.s3.eu-west-1.amazonaws.com/productimage/Web-Homepage-graphic-1.jpg'
    assert order_item.price == 1
    assert order_item.total_price == 1
    assert isinstance(order_item.product_variant, ProductVariant)
    assert order_item.product_variant.id == 1
    assert order_item.product_variant.price == 0
    assert order_item.product_variant.stocks == [
        '/v4/product_variant_stocks/fulfilmentPoint=1;productVariant=1'
    ]
    assert isinstance(order_item.product, Product)
    assert order_item.product.id == 1
    assert order_item.product.status == 'ACTIVE'
    assert order_item.product.reference == 'Plant'
    assert order_item.product.images == []
    assert len(order_item.product.categories) == 1
    assert isinstance(order_item.product.categories[0], ProductCategory)
    assert order_item.product.categories[0].id == 1
    assert order_item.product.categories[0].reference == 'Kartiks product Category'
    assert order_item.product.categories[0].position == 1
    assert isinstance(order_item.fulfilment_point, FulfilmentPoint)
    assert order_item.fulfilment_point.id == 1
    assert order_item.fulfilment_point.status == 'ACTIVE'
    assert order_item.fulfilment_point.reference == 'kartiks Collection point'
    assert order_item.fulfilment_point.image_url == 'https://apiv3.s3.eu-west-1.amazonaws.com/fulfilmentpoint/German-Shepherd-Puppy-Fetch.jpg'
    assert order_item.fulfilment_point.map_image_url == 'https://apiv3.s3.eu-west-1.amazonaws.com/fulfilmentpoint/british_summer_time_map_bigger.jpg'
    assert order_item.fulfilment_point.lat == '38.026641'
    assert order_item.fulfilment_point.long == '38.026641'
    assert order_item.fulfilment_point.type == 'COLLECTION'
    assert order_item.fulfilment_point.position == 0
    assert len(order_item.fulfilment_point.translations) == 1
    assert isinstance(order_item.fulfilment_point.translations[0], FulfilmentPointTranslation)
    assert order_item.fulfilment_point.translations[0].id == 1
    assert order_item.fulfilment_point.translations[0].language == 'en'
    assert order_item.fulfilment_point.translations[0].description == 'Description'
    assert order_item.fulfilment_point.translations[0].title == 'Fulfilment point 1'
    assert order_item.fulfilment_point.translations[0].collection_note == 'Fulfillment note'
    assert len(order_item.fulfilment_point.categories) == 1
    assert isinstance(order_item.fulfilment_point.categories[0], FulfilmentPointCategory)
    assert order_item.fulfilment_point.categories[0].id == 1
    assert order_item.fulfilment_point.categories[0].status == 'ACTIVE'
    assert len(order_item.fulfilment_point.categories[0].translations) == 1
    assert isinstance(order_item.fulfilment_point.categories[0].translations[0], FulfilmentPointCategoryTranslation)
    assert order_item.fulfilment_point.categories[0].translations[0].id == 1
    assert order_item.fulfilment_point.categories[0].translations[0].language == 'en'
    assert order_item.fulfilment_point.categories[0].translations[0].title == 'Kartiks Collection Point Category'
    assert isinstance(order.user, User)
    assert order.user.id == 274814
    assert order.user.email == 'test@gmail.com'
    assert order.user.token == 'd290fab6-9f89-413d-a89b-9dd5dbcd8eb4'
    assert order.user.user_info.first_name == 'string'
    assert order.user.user_info.last_name == 'string'
    assert order.user.user_info.gender == 'MALE'
    assert order.user.user_info.phone == '+4712345678'
    assert order.user.user_info.dob == datetime(2020, 4, 30, 0, 0, tzinfo=timezone(timedelta(0), '+0000'))
    assert isinstance(order.app, App)
    assert isinstance(order.app.currency, Currency)
    assert order.app.currency.iso_code == 'GBP'
    assert order.app.currency.title == 'Pound'
    assert order.app.currency.sign == '£'



from livestyled.models.venue import Venue


class FulfilmentPointCategoryTranslation:
    def __init__(
            self,
            id,
            language,
            title
    ):
        self.id = id
        self.language = language
        self.title = title


class FulfilmentPointTranslation:
    def __init__(
            self,
            id,
            language,
            title,
            description,
            collection_note
    ):
        self.id = id
        self.language = language
        self.title = title
        self.description = description
        self.collection_note = collection_note


class FulfilmentPointCategory:
    def __init__(
            self,
            id,
            status,
            translations,
    ):
        self.id = id
        self.status = status

        if translations:
            for translation in translations:
                self.translations = []
                if isinstance(translation, FulfilmentPointCategoryTranslation):
                    self.translations.append(translation)
                elif isinstance(translation, dict):
                    self.translations.append(FulfilmentPointCategoryTranslation(**translation))
        else:
            self.translations = []


class FulfilmentPoint:
    def __init__(
            self,
            id,
            status,
            image_url,
            map_image_url,
            lat,
            long,
            type,
            position,
            reference,
            translations,
            categories,
            venue
    ):
        self.id = id
        self.status = status
        self.image_url = image_url
        self.map_image_url = map_image_url
        self.lat = lat
        self.long = long
        self.type = type
        self.position = position
        self.reference = reference

        if translations:
            for translation in translations:
                self.translations = []
                if isinstance(translation, FulfilmentPointTranslation):
                    self.translations.append(translation)
                elif isinstance(translation, dict):
                    self.translations.append(FulfilmentPointTranslation(**translation))
        else:
            self.translations = []

        if categories:
            for category in categories:
                self.categories = []
                if isinstance(category, FulfilmentPointCategory):
                    self.categories.append(category)
                elif isinstance(category, dict):
                    self.categories.append(FulfilmentPointCategory(**category))
        else:
            self.categories = []

        if venue:
            if isinstance(venue, Venue):
                self.venue = venue
            elif isinstance(venue, dict):
                self.venue = Venue(**venue)
            elif isinstance(venue, (str, int)):
                self.venue = Venue.placeholder(id=venue)
        else:
            self.venue = None

    @classmethod
    def placeholder(cls, id):
        return cls(
            id,
            status=None,
            image_url=None,
            map_image_url=None,
            lat=None,
            long=None,
            type=None,
            position=None,
            reference=None,
            translations=None,
            categories=None,
            venue=None,
        )



class DeviceConsent:
    def __init__(
            self,
            id: int,
            location_capture: bool,
            location_granular: bool,
            camera: bool,
            calendar: bool,
            photo_sharing: bool,
            push_notification: bool,
            created_at,
            updated_at
    ):
        self.id = id
        self.location_capture = location_capture
        self.location_granular = location_granular
        self.camera = camera
        self.calendar = calendar
        self.photo_sharing = photo_sharing
        self.push_notification = push_notification
        self.created_at = created_at
        self.updated_at = updated_at



from marshmallow import EXCLUDE, fields, Schema

from livestyled.models.push_broadcast import PushBroadcast
from livestyled.schemas.fields import RelatedResourceLinkField
from livestyled.schemas.push_consent import PushConsentSchema


class PushBroadcastSchema(Schema):
    class Meta:
        unknown = EXCLUDE
        api_type = 'push_broadcasts'
        url = 'v4/push_broadcasts'
        publish_url = 'v4/push_broadcasts/publish'
        model = PushBroadcast

    id = fields.Int(required=False, allow_none=False)
    title = fields.String(data_key='title')
    message = fields.String(data_key='message')
    deep_link = fields.String(data_key='deepLink')
    ttl = fields.Int()
    message_id = fields.String(data_key='messageId')
    delivered = fields.Int()
    updated_at = fields.AwareDateTime(data_key='updatedAt', allow_none=True)
    created_at = fields.AwareDateTime(data_key='createdAt', allow_none=True)
    push_consent_id = RelatedResourceLinkField(schema=PushConsentSchema, data_key='pushConsent')



from marshmallow import EXCLUDE, fields, Schema

from livestyled.models.league_table import LeagueTable, LeagueTableGroup
from livestyled.schemas.competition import CompetitionSchema
from livestyled.schemas.fields import RelatedResourceLinkField
from livestyled.schemas.season import SeasonSchema
from livestyled.schemas.team import TeamSchema


class LeagueTableGroupSchema(Schema):
    class Meta:
        unknown = EXCLUDE
        api_type = 'league_table_groups'
        url = 'v4/league_table_groups'
        model = LeagueTableGroup

    id = fields.Int(required=False, allow_none=False)
    reference = fields.String(required=True, allow_none=True)
    title = fields.String(required=True, allow_none=True)


class LeagueTableSchema(Schema):
    class Meta:
        unknown = EXCLUDE
        api_type = 'league_tables'
        url = 'v4/league_tables'
        model = LeagueTable

    id = fields.Int(required=False, allow_none=False)
    team_id = RelatedResourceLinkField(schema=TeamSchema, data_key='team')
    external_id = fields.String(required=True, allow_none=True, data_key='externalId')
    featured_team = fields.Boolean(required=True, allow_none=False, data_key='featuredTeam')
    position = fields.Int(missing=0)
    played = fields.Int(missing=0)
    goals_for = fields.Int(data_key='goalsFor', missing=0)
    goals_against = fields.Int(data_key='goalsAgainst', missing=0)
    start_day_position = fields.Int(data_key='startDayPosition', missing=0)
    won = fields.Int(missing=0)
    lost = fields.Int(missing=0)
    drawn = fields.Int(missing=0)
    goal_difference = fields.Int(data_key='goalDifference', missing=0)
    points = fields.Int(missing=0)
    competition_id = RelatedResourceLinkField(schema=CompetitionSchema, data_key='competition')
    season_id = RelatedResourceLinkField(schema=SeasonSchema, data_key='season')
    group_id = RelatedResourceLinkField(schema=LeagueTableGroupSchema, data_key='group')



class Cohort:
    def __init__(
            self,
            title,
            external_id=None
    ):
        self.id = None
        self.title = title
        self.external_id = external_id

    @classmethod
    def placeholder(
            cls,
            id
    ):
        cohort = cls(
            title=None,
            external_id=None,
        )
        cohort.id = id
        return cohort

    def __repr__(self):
        return '<Cohort(id={self.id!r}, title={self.title!r}, external_id={self.external_id!r})>'.format(self=self)

    def __hash__(self):
        return hash((self.id, self.title, self.external_id))

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.id == other.id and self.title == other.title and self.external_id == other.external_id

    def diff(self, other):
        differences = {}
        fields = (
            'title', 'external_id'
        )
        for field in fields:
            if getattr(self, field) != getattr(other, field):
                differences[field] = getattr(self, field)
        return differences



from livestyled.models.device import Device


class DeviceToken:
    def __init__(
            self,
            id: int or None,
            provider: str,
            provider_token: str,
            payload,
            device_id: str or None = None
    ):
        self._id = id
        self.provider = provider
        self.provider_token = provider_token
        self.payload = payload
        self._device = Device.placeholder(id=device_id)

    @property
    def id(self):
        return self._id

    @classmethod
    def create_new(
            cls,
            provider,
            provider_token,
            payload,
            device
    ):
        token = DeviceToken(
            id=None,
            provider=provider,
            provider_token=provider_token,
            payload=payload,
            device_id=None
        )
        if isinstance(device, (str, int)):
            device = Device.placeholder(id=device)
        token._device = device
        return token

    def __repr__(self):
        return '<DeviceToken(id={self.id!r}, provider={self.provider!r})>'.format(self=self)

    def diff(self, other):
        differences = {}
        fields = (
            'provider', 'provider_token', 'payload',
        )
        for field in fields:
            if getattr(self, field) != getattr(other, field):
                differences[field] = getattr(self, field)
        return differences

    @property
    def device_id(self):
        return self._device.id

    @property
    def device(self):
        return self._device



import json
from typing import Dict, Generator, List, Type

from marshmallow import Schema
import requests

from livestyled.schemas import (
    AppSchema,
    DeviceSchema,
    DeviceTokenSchema,
    EventCategorySchema,
    EventIntegrationSchema,
    EventSchema,
    EventStageSchema,
    NewsSchema,
    PushConsentSchema,
    TeamSchema,
    UserSchema,
)

CONTENT_TYPE = 'application/ld+json'
PATCH_CONTENT_TYPE = 'application/merge-patch+json'


class LiveStyledAPIClient:
    def __init__(
            self,
            api_domain: str,
            api_key: str
    ):
        self._api_domain = api_domain
        self._api_key = api_key

    def _api_get(
            self,
            endpoint: str,
            params: Dict or None = None
    ) -> Dict:
        if not params:
            params = {}
        response = requests.get(
            'https://{}/{}'.format(
                self._api_domain,
                endpoint,
            ),
            params=params,
            headers={
                'Content-Type': CONTENT_TYPE,
                'x-api-key': self._api_key
            }
        )
        response.raise_for_status()
        return response.json()

    def _api_get_paginated(
            self,
            endpoint: str,
            params: Dict or None = None,
    ) -> Generator[Dict, None, None]:
        if params:
            response = self._api_get(endpoint, params)
        else:
            response = self._api_get(endpoint)
        resources = response['hydra:member']
        for resource in resources:
            yield resource
        next_page_links = response.get('hydra:view', None)
        if not next_page_links:
            return
        next_page_url = next_page_links.get('hydra:next', False)
        while next_page_url:
            next_page_url = next_page_url.replace('//', '/').lstrip('/')
            response = self._api_get(next_page_url)
            resources = response['hydra:member']
            next_page_url = response['hydra:view'].get('hydra:next', False)
            for resource in resources:
                yield resource

    def _api_patch(
            self,
            endpoint: str,
            data: dict
    ) -> Dict:
        response = requests.patch(
            'https://{}/{}'.format(
                self._api_domain,
                endpoint,
            ),
            headers={
                'Content-Type': PATCH_CONTENT_TYPE,
                'x-api-key': self._api_key
            },
            data=json.dumps(data)
        )
        response.raise_for_status()
        return response.json()

    def _api_post(
            self,
            endpoint: str,
            data: dict,
            content_type_override: str or None = None
    ) -> Dict:
        content_type = CONTENT_TYPE
        if content_type_override:
            content_type = content_type_override
        response = requests.post(
            'https://{}/{}'.format(
                self._api_domain,
                endpoint,
            ),
            headers={
                'Content-Type': content_type,
                'x-api-key': self._api_key
            },
            data=json.dumps(data)
        )
        if response.status_code >= 400:
            print(response.json())
        response.raise_for_status()
        return response.json()

    def _api_delete(
            self,
            endpoint: str,
            id: str
    ) -> None:
        requests.delete(
            'https://{}/{}/{}'.format(
                self._api_domain,
                endpoint,
                id
            ),
            headers={
                'Content-Type': CONTENT_TYPE,
                'x-api-key': self._api_key
            },
        )
        return

    def _get_resource(
            self,
            resource_id: int or str,
            schema: Type[Schema]
    ):
        resource_data = self._api_get(
            '{}/{}'.format(
                schema.Meta.url,
                resource_id
            )
        )
        resource = schema().load(resource_data)
        return resource

    def _get_resources(
            self,
            schema: Type[Schema],
            params: Dict or None = None,
    ) -> Generator[Dict, None, None]:
        data_generator = self._api_get_paginated(
            schema.Meta.url,
            params
        )
        for resource_data in data_generator:
            yield schema().load(resource_data)

    def get_app(
            self,
            app_id: int or str,
    ) -> Dict:
        return self._get_resource(
            app_id,
            AppSchema,
        )

    def get_apps(
            self,
    ) -> Generator[Dict, None, None]:
        return self._get_resources(AppSchema)

    def get_device(
            self,
            device_id: int or str,
    ) -> Dict:
        return self._get_resource(
            device_id,
            DeviceSchema,
        )

    def get_devices(
            self,
            token: str or None = None,
            tokens: List[str] or None = None,
            status: str or None = None,
            statuses: List[str] or None = None
    ) -> Generator[Dict, None, None]:
        filter_params = {}
        if token:
            filter_params['token'] = token
        if tokens:
            filter_params['token[]'] = tokens
        if token and tokens:
            # TODO error as mutually exclusive
            pass
        if status:
            filter_params['status'] = status
        if statuses:
            filter_params['status[]'] = statuses
        if status and statuses:
            # TODO error as mutually exclusive
            pass

        return self._get_resources(
            DeviceSchema,
            params=filter_params
        )

    def update_device(
            self,
            device_id: str,
            attributes: Dict,
    ) -> Dict:
        attributes_to_update = list(attributes.keys())
        update_payload = DeviceSchema(only=attributes_to_update).dump(attributes)
        updated_device = self._api_patch(
            '{}/{}'.format(DeviceSchema.Meta.url, device_id),
            update_payload
        )
        return DeviceSchema().load(updated_device)

    # def get_device_consent(
    #         self,
    #         consent_id: int or str,
    # ) -> Dict:
    #     return self._get_resource(
    #         consent_id,
    #         DeviceConsentSchema,
    #     )
    #
    # def get_device_consents(
    #         self,
    # ) -> Generator[Dict, None, None]:
    #     return self._get_resources(
    #         DeviceConsentSchema,
    #     )

    def get_device_token(
            self,
            token_id: int or str,
    ) -> Dict:
        return self._get_resource(
            token_id,
            DeviceTokenSchema,
        )

    def get_device_tokens(
            self,
            provider: str or None = None,
            providers: List[str] or None = None,
            device: str or None = None,
            devices: List[str] or None = None
    ) -> Generator[Dict, None, None]:
        filter_params = {}
        if provider:
            filter_params['provider'] = provider
        if providers:
            filter_params['provider[]'] = providers
        if provider and providers:
            # TODO error as mutually exclusive
            pass
        if device:
            filter_params['device'] = device
        if devices:
            filter_params['device[]'] = device
        if device and devices:
            # TODO error as mutually exclusive
            pass
        return self._get_resources(
            DeviceTokenSchema,
            params=filter_params,
        )

    def get_event(
            self,
            event_id: int or str,
    ) -> Dict:
        return self._get_resource(
            event_id,
            EventSchema,
        )

    def get_events(
            self,
    ) -> Generator[Dict, None, None]:
        return self._get_resources(EventSchema)

    def get_event_category(
            self,
            category_id: int or str,
    ) -> Dict:
        return self._get_resource(
            category_id,
            EventCategorySchema,
        )

    def get_event_categories(
            self,
    ) -> Generator[Dict, None, None]:
        return self._get_resources(EventCategorySchema)

    def get_event_integration(
            self,
            integration_id: int or str,
    ) -> Dict:
        return self._get_resource(
            integration_id,
            EventIntegrationSchema,
        )

    def get_event_integrations(
            self,
    ) -> Generator[Dict, None, None]:
        return self._get_resources(EventIntegrationSchema)

    def get_event_stage(
            self,
            stage_id: int or str,
    ) -> Dict:
        return self._get_resource(
            stage_id,
            EventStageSchema,
        )

    def get_event_stages(
            self,
    ) -> Generator[Dict, None, None]:
        return self._get_resources(EventStageSchema)

    def get_push_consent(
            self,
            consent_id: int or str,
    ) -> Dict:
        return self._get_resource(
            consent_id,
            PushConsentSchema,
        )

    def get_push_consents(
            self,
    ) -> Generator[Dict, None, None]:
        return self._get_resources(PushConsentSchema)

    def get_user(
            self,
            user_id: int or str,
    ) -> Dict:
        return self._get_resource(
            user_id,
            UserSchema,
        )

    def get_users(
            self,
    ) -> Generator[Dict, None, None]:
        return self._get_resources(UserSchema)

    def get_teams(
            self,
            external_id: str or None = None,
    ) -> Generator[Dict, None, None]:
        filter_params = {}
        if external_id:
            filter_params['externalId'] = external_id

        return self._get_resources(
            TeamSchema,
            params=filter_params
        )

    def update_team(
            self,
            team_id: str,
            attributes: Dict,
    ) -> Dict:
        attributes_to_update = list(attributes.keys())
        update_payload = TeamSchema(only=attributes_to_update).dump(attributes)
        updated_team = self._api_patch(
            '{}/{}'.format(TeamSchema.Meta.url, team_id),
            update_payload
        )
        return TeamSchema().load(updated_team)

    def create_team(
            self,
            attributes: Dict,
    ) -> Dict:
        payload = TeamSchema().dump(attributes)
        new_team = self._api_post(
            '{}'.format(TeamSchema.Meta.url),
            payload
        )
        return TeamSchema().load(new_team)

    def get_news(
            self,
            external_id: str or None = None,
    ) -> Generator[Dict, None, None]:
        filter_params = {}
        if external_id:
            filter_params['externalId'] = external_id

        return self._get_resources(
            NewsSchema,
            params=filter_params
        )

    def create_news(
            self,
            attributes: Dict,
    ) -> Dict:
        payload = NewsSchema().dump(attributes)
        new_team = self._api_post(
            '{}'.format(NewsSchema.Meta.url),
            payload
        )
        return NewsSchema().load(new_team)



from livestyled.models.fulfilment_point import FulfilmentPoint


class ProductVariant:
    def __init__(
            self,
            id,
            price,
            stocks
    ):
        self.id = id
        self.price = price
        self.stocks = stocks

    @classmethod
    def placeholder(cls, id):
        return cls(
            id,
            price=None,
            stocks=None,
        )


class ProductTranslation:
    def __init__(
            self,
            id,
            language,
            title,
            description,
            created_at,
            updated_at
    ):
        self.id = id
        self.language = language
        self.title = title
        self.description = description
        self.created_at = created_at
        self.updated_at = updated_at


class ProductImage:
    def __init__(
            self,
            id,
            product,
            position,
            created_at,
            updated_at,
            external_id,
            image_url
    ):
        self.id = id
        self.position = position
        self.created_at = created_at
        self.updated_at = updated_at
        self.external_id = external_id
        self.image_url = image_url
        self.product = None
        if product:
            if isinstance(product, Product):
                self.product = product
            elif isinstance(product, dict):
                self.product = Product(**product)
            elif isinstance(product, int):
                self.product = Product.placeholder(id=product)


class ProductCategory:
    def __init__(
            self,
            id,
            reference,
            position
    ):
        self.id = id
        self.reference = reference
        self.position = position


class Product:
    def __init__(
            self,
            id,
            status,
            reference,
            images,
            categories,
            translations,
            modifier_lists,
            external_id,
            holding_time,
            reconciliation_group,
            fulfilment_points,
            variants,
            core_product_category
    ):
        self.id = id
        self.status = status
        self.reference = reference
        self.modifier_lists = modifier_lists
        self.external_id = external_id
        self.holding_time = holding_time
        self.reconciliation_group = reconciliation_group
        self.core_product_category = core_product_category
        if images:
            self.images = []
            for image in images:
                if isinstance(image, ProductImage):
                    self.images.append(image)
                elif isinstance(image, dict):
                    self.images.append(ProductImage(**image))
        else:
            self.images = []

        if categories:
            self.categories = []
            for category in categories:
                if isinstance(category, ProductCategory):
                    self.categories.append(category)
                elif isinstance(category, dict):
                    self.categories.append(ProductCategory(**category))
        else:
            self.categories = []

        if translations:
            self.translations = []
            for translation in translations:
                if isinstance(translation, ProductTranslation):
                    self.translations.append(translation)
                elif isinstance(translation, dict):
                    self.translations.append(ProductTranslation(**translation))
        else:
            self.translations = []

        if fulfilment_points:
            self.fulfilment_points = []
            for fulfilment_point in fulfilment_points:
                if isinstance(fulfilment_point, FulfilmentPoint):
                    self.fulfilment_points.append(fulfilment_point)
                elif isinstance(fulfilment_point, dict):
                    self.fulfilment_points.append(FulfilmentPoint(**fulfilment_point))
                elif isinstance(fulfilment_point, int):
                    self.fulfilment_points.append(FulfilmentPoint.placeholder(id=fulfilment_point))

        if variants:
            self.variants = []
            for variant in variants:
                if isinstance(variant, ProductVariant):
                    self.variants.append(variant)
                elif isinstance(variant, dict):
                    self.variants.append(ProductVariant(**variant))
                elif isinstance(variant, int):
                    self.variants.append(ProductVariant.placeholder(id=variant))

    @classmethod
    def placeholder(
            cls,
            id
    ):
        return cls(
            id=id,
            status=None,
            reference=None,
            images=None,
            categories=None,
            translations=None,
            external_id=None,
            variants=None,
            fulfilment_points=None,
            holding_time=None,
            reconciliation_group=None,
            modifier_lists=None,
            core_product_category=None
        )



from marshmallow import EXCLUDE, fields, Schema

from livestyled.models.app import App, Currency


class CurrencySchema(Schema):
    class Meta:
        unknown = EXCLUDE
        model = Currency
        api_type = 'currencies'
        url = 'v4/currencies'

    id = fields.Int(missing=None)
    title = fields.String(missing=None)
    iso_code = fields.String(missing=None, data_key='isoCode')
    sign = fields.String(missing=None)


class AppSchema(Schema):
    class Meta:
        unknown = EXCLUDE
        api_type = 'apps'
        url = 'v4/apps'
        model = App

    id = fields.Int(missing=None)
    email = fields.Email(allow_none=True, missing=None)
    roles = fields.List(fields.String(), allow_none=True, missing=None)
    password = fields.String(missing=None)
    salt = fields.String(allow_none=True, missing=None)
    username = fields.String(missing=None)
    code = fields.String(missing=None)
    token = fields.String(missing=None)
    name = fields.String(missing=None)
    api_keys = fields.List(fields.String(), data_key='apiKeys', missing=None)
    timezone = fields.String(missing=None)
    google_api_key = fields.String(data_key='googleApiKey', allow_none=True, missing=None)
    payment_gateway = fields.String(data_key='paymentGateway', missing=None)
    venues = fields.List(fields.Inferred(), missing=None)  # TODO
    sender_email = fields.Email(data_key='senderEmail', missing=None)
    deeplink_namespace = fields.String(data_key='deeplinkNamespace', missing=None)
    merchant_accounts = fields.List(fields.Inferred, data_key='merchantAccounts', allow_none=True, missing=None)
    title = fields.String(missing=None)
    status = fields.String(missing=None)
    currency = fields.Nested(CurrencySchema, missing=None)
    cohorts = fields.List(fields.Inferred, missing=None)



from marshmallow import EXCLUDE, fields, Schema

from livestyled.models.booking import Booking
from livestyled.schemas.device import DeviceSchema
from livestyled.schemas.event import EventSchema
from livestyled.schemas.fields import RelatedResourceLinkField
from livestyled.schemas.user import UserSchema


class BookingSchema(Schema):
    class Meta:
        unknown = EXCLUDE
        model = Booking
        api_type = 'bookings'
        url = 'v4/bookings'

    id = fields.Int()
    title = fields.String()
    device_id = RelatedResourceLinkField(schema=DeviceSchema, data_key='device')
    user_id = RelatedResourceLinkField(schema=UserSchema, data_key='user')
    event_id = RelatedResourceLinkField(schema=EventSchema, data_key='event')
    updated_at = fields.AwareDateTime(data_key='updatedAt', allow_none=True)
    created_at = fields.AwareDateTime(data_key='createdAt', allow_none=True)
    action = fields.String()
    type = fields.String()



from marshmallow import EXCLUDE, fields, Schema

from livestyled.models.ticket_integration import TicketIntegration


class TicketIntegrationSchema(Schema):
    class Meta:
        unknown = EXCLUDE
        api_type = 'ticket_integrations'
        url = 'v4/ticket_integrations'
        model = TicketIntegration

    id = fields.Int()
    label = fields.String(required=False, missing=None)
    name = fields.String(required=False, missing=None)
    adapter = fields.String(required=False, missing=None)
    endpoint_url = fields.String(required=False, missing=None, data_key='endpointUrl')
    config_payload = fields.String(required=False, missing=None, data_key='configPayload')
    auth_required = fields.Boolean(required=False, missing=None, data_key='authRequired')
    module = fields.String(required=False, missing=None)
    login_request = fields.String(required=False, missing=None, data_key='loginRequest')
    default = fields.Boolean(required=False, missing=None, data_key='default')
    can_share = fields.Boolean(required=False, missing=None, data_key='canShare')



from marshmallow import EXCLUDE, fields, Schema

from livestyled.models.news import News


class NewsSchema(Schema):
    class Meta:
        unknown = EXCLUDE
        api_type = 'news'
        url = 'v4/news'
        model = News

    class NewsMediaSchema(Schema):
        class Meta:
            unknown = EXCLUDE

        type = fields.String(required=True, allow_none=False)
        url = fields.String(required=True, allow_none=False)
        thumbnail_url = fields.String(required=False, allow_none=False, missing=None, data_key='thumbnailUrl')

    id = fields.Int()
    external_id = fields.String(data_key='externalId')
    headline = fields.String(data_key='headline', missing=None)
    image_url = fields.String(data_key='imageUrl', missing=None)
    title = fields.String(data_key='title')
    media = fields.Nested(NewsMediaSchema, missing=None)
    url = fields.String(allow_none=True)
    updated_at = fields.AwareDateTime(data_key='updatedAt', allow_none=False, load_only=True)
    published_at = fields.AwareDateTime(data_key='publishedAt', allow_none=False)
    author = fields.String(required=False, missing=None)



schema_map = {}


def _generate_schema_map():
    global schema_map
    from livestyled import schemas
    for schema in schemas.__all__:
        schema_map[schema.Meta.model.__name__] = schema


def create_resource_from_data(resource_model, data):
    if not schema_map:
        _generate_schema_map()
    schema = schema_map[resource_model.__name__]
    d = schema().load(data)
    return resource_model(**d)



from marshmallow import EXCLUDE, fields, Schema

from livestyled.models.season import Season


class SeasonSchema(Schema):
    class Meta:
        unknown = EXCLUDE
        api_type = 'seasons'
        url = 'v4/seasons'
        model = Season

    id = fields.Int()
    label = fields.String()
    name = fields.String()
    is_current = fields.Boolean(data_key='isCurrent')
    external_id = fields.String(data_key='externalId', missing=None)



from marshmallow import EXCLUDE, fields, Schema
from marshmallow_polyfield import PolyField

from livestyled.models.ticket import Ticket
from livestyled.schemas.event import EventSchema
from livestyled.schemas.fields import RelatedResourceField, RelatedResourceLinkField
from livestyled.schemas.ticket_integration import TicketIntegrationSchema
from livestyled.schemas.user import UserSchema
from livestyled.schemas.venue import VenueSchema


def parent_ticket_selector(parent_ticket, ticket):
    if isinstance(parent_ticket, str):
        return RelatedResourceLinkField(schema=Ticket, required=False, missing=None, data_key='parentTicket')
    elif isinstance(parent_ticket, dict):
        return fields.Nested(TicketSchema)


class TicketSchema(Schema):
    class Meta:
        unknown = EXCLUDE
        api_type = 'tickets'
        url = 'v4/tickets'
        model = Ticket

    id = fields.Int()
    external_ticket_id = fields.String(data_key='externalTicketId')
    seat = fields.String(required=False, missing=None)
    qr_code_url = fields.String(data_key='qrCodeUrl', required=False, missing=None)
    session_date = fields.AwareDateTime(data_key='sessionDate', allow_none=False)
    title = fields.String(required=None, missing=False)
    external_event_id = fields.String(data_key='eventUid', required=False, missing=None)
    barcode = fields.String(data_key='barCode', required=False, missing=None)
    sector_name = fields.String(data_key='sectorName', required=False, missing=None)
    venue_name = fields.String(data_key='venueName', required=False, missing=None)
    venue_room = fields.String(data_key='venueRoom', required=False, missing=None)
    client_name = fields.String(data_key='clientName', required=False, missing=None)
    premium = fields.Boolean()
    client_email = fields.String(data_key='clientEmail', required=False, missing=None)
    price = fields.Int(required=False, missing=None)
    share_link = fields.String(data_key='shareLink', required=False, missing=None)
    external_customer_ref = fields.String(data_key='externalCustomerRef', required=False, missing=None)
    entrance = fields.String(required=False, missing=None)
    section = fields.String(required=False, missing=None)
    row = fields.String(required=False, missing=None)
    status = fields.String(required=False, missing=None)
    price_code = fields.String(data_key='priceCode', required=False, missing=None)
    created_at = fields.AwareDateTime(data_key='createdAt', allow_none=False)
    updated_at = fields.AwareDateTime(data_key='updatedAt', allow_none=False)
    user_id = RelatedResourceLinkField(schema=UserSchema, required=False, missing=None, data_key='user')
    can_share = fields.Boolean(data_key='canShare', allow_none=False, required=False, missing=False)
    share_code = fields.String(data_key='shareCode', allow_none=True, required=False, missing=None)
    sharer_email = fields.String(data_key='sharerEmail', allow_none=True, required=False, missing=None)
    redeemer_email = fields.String(data_key='redeemerEmail', allow_none=True, required=False, missing=None)
    redeemed_at = fields.AwareDateTime(data_key='redeemedAt', required=False, missing=None)
    shared_at = fields.AwareDateTime(data_key='sharedAt', required=False, missing=None)
    sharer_id = RelatedResourceLinkField(schema=UserSchema, required=False, missing=None, data_key='sharer')
    redeemer_id = RelatedResourceLinkField(schema=UserSchema, required=False, missing=None, data_key='redeemer')
    parent_ticket = PolyField(
        deserialization_schema_selector=parent_ticket_selector,
        data_key='parentTicket',
        required=False,
        missing=None,
        allow_none=True
    )
    legal_short_text = fields.String(data_key='legalShortText', required=False, allow_none=True, missing=None)
    legal_long_text = fields.String(data_key='legalLongText', required=False, allow_none=True, missing=None)
    map_url = fields.String(data_key='mapUrl', required=False, allow_none=True, missing=None)
    map_image_url = fields.String(data_key='mapImageUrl', required=False, allow_none=True, missing=None)
    ticket_integration = RelatedResourceField(schema=TicketIntegrationSchema, required=False, missing=None, data_key='ticketIntegration')
    event = RelatedResourceLinkField(schema=EventSchema, required=False, missing=None)
    venue = RelatedResourceLinkField(schema=VenueSchema, required=False, missing=None)



from livestyled.models.push_consent import PushConsent


class DevicePushConsent:
    def __init__(
            self,
            id: int,
            consent: bool,
            push_consent_id: str
    ):
        self.id = id
        self.consent = consent
        self._push_consent = PushConsent.placeholder(id=push_consent_id)

    @property
    def push_consent(self):
        return self._push_consent



from livestyled.models.push_consent import PushConsent


class PushBroadcast:
    def __init__(
            self,
            id: int or None,
            title: str or None,
            message,
            message_id,
            updated_at,
            created_at,
            delivered=None,
            push_consent_id=None,
            deep_link=None,
            ttl=None,
    ):
        self.id = id
        self.title = title
        self.message = message
        self.deep_link = deep_link
        self.ttl = ttl
        self.message_id = message_id
        self.delivered = delivered
        self.updated_at = updated_at
        self.created_at = created_at
        if push_consent_id:
            self._push_consent = PushConsent.placeholder(id=push_consent_id)
        else:
            self._push_consent = None

    @classmethod
    def create_new(
            cls,
            title,
            message,
            deep_link,
            ttl=0,
            message_id=None,
            delivered=0,
            updated_at=None,
            created_at=None,
            push_consent=None
    ):
        push_broadcast = PushBroadcast(
            id=None,
            title=title,
            message=message,
            deep_link=deep_link,
            ttl=ttl,
            message_id=message_id,
            delivered=delivered,
            updated_at=updated_at,
            created_at=created_at,
            push_consent_id=None
        )
        push_broadcast._push_consent = push_consent
        return push_broadcast

    @classmethod
    def placeholder(
            cls,
            id
    ):
        return cls(
            id=id,
            title=None,
            message=None,
            deep_link=None,
            ttl=None,
            message_id=None,
            delivered=None,
            updated_at=None,
            created_at=None,
            push_consent_id=None
        )

    @property
    def push_consent_id(self):
        return self._push_consent.id

    @property
    def push_consent(self):
        return self._push_consent

    def diff(self, other):
        differences = {}
        fields = (
            'title', 'message', 'deep_link', 'ttl', 'message_id', 'delivered',
            'updated_at', 'created_at', 'push_consent_id'
        )
        for field in fields:
            if getattr(self, field) != getattr(other, field):
                differences[field] = getattr(self, field)
        return differences



import os

import pytest

from livestyled.models.team import Team
from livestyled.resource_client import LiveStyledResourceClient
from livestyled.tests.utils import configure_mock_responses

FIXTURES_DIR = os.path.join(os.path.dirname(__file__), 'fixtures')
TEST_API_DOMAIN = 'test.livestyled.com'
CONTENT_TYPE = 'application/ld+json'


def test_get_team_by_external_id(requests_mock):
    mock_responses = (
        ('GET', 'https://' + TEST_API_DOMAIN + '/v4/teams?externalId=t3', 'mock_responses/ls_api/teams_external_id_t3.json', 200),
    )
    configure_mock_responses(requests_mock, mock_responses, FIXTURES_DIR, CONTENT_TYPE)

    resource_client = LiveStyledResourceClient(TEST_API_DOMAIN, 'bar')
    teams = list(resource_client.get_teams(external_id='t3'))
    assert teams
    team = teams[0]
    assert team.name == 'Arsenal'
    assert team.short_name == 'ARS'
    assert team.light_crest_url == 'https://apiv3-dev.s3.eu-west-1.amazonaws.com/27/team/arsenal-300x300.png'
    assert team.dark_crest_url == 'https://apiv3-dev.s3.eu-west-1.amazonaws.com/27/team/arsenal-300x300.png'
    assert team.external_id == 't3'


def test_get_team_by_id(requests_mock):
    mock_responses = (
        ('GET', 'https://' + TEST_API_DOMAIN + '/v4/teams/6', 'mock_responses/ls_api/team_6.json', 200),
    )
    configure_mock_responses(requests_mock, mock_responses, FIXTURES_DIR, CONTENT_TYPE)

    resource_client = LiveStyledResourceClient(TEST_API_DOMAIN, 'bar')
    team = resource_client.get_team(6)
    assert team.name == 'LiveStyled'
    assert team.short_name == 'LS'
    assert team.light_crest_url == 'https://apiv3.s3.eu-west-1.amazonaws.com/team/Logo_White_nontext.png'
    assert team.dark_crest_url == 'https://apiv3.s3.eu-west-1.amazonaws.com/team/Logo_Purple_notext.png'
    assert team.external_id is None


def test_create_team(requests_mock):
    team = Team.create_new(
        external_id='testteam1',
        name='Test Team',
        short_name='TT',
        light_crest_url='http://www.images.com/image1.jpg',
        dark_crest_url='http://www.images.com/image2.jpg'
    )
    mock_responses = (
        ('POST', 'https://' + TEST_API_DOMAIN + '/v4/teams', 'mock_responses/ls_api/create_team.json', 200),
    )
    configure_mock_responses(requests_mock, mock_responses, FIXTURES_DIR, CONTENT_TYPE)

    resource_client = LiveStyledResourceClient(TEST_API_DOMAIN, 'bar')
    created_team = resource_client.create_team(team)
    assert created_team
    assert created_team.id == 999

    for request in requests_mock.request_history:
        if request.method.upper() == 'POST' and request.url == 'https://' + TEST_API_DOMAIN + '/v4/teams':
            actual_payload = request.json()

            assert actual_payload == {
                'name': 'Test Team',
                'shortName': 'TT',
                'lightCrestUrl': 'http://www.images.com/image1.jpg',
                'darkCrestUrl': 'http://www.images.com/image2.jpg',
                'externalId': 'testteam1'
            }


def test_create_team_with_id(requests_mock):
    team = Team(
        id=1,
        external_id='testteam1',
        name='Test Team',
        short_name='TT',
        light_crest_url='http://www.images.com/image1.jpg',
        dark_crest_url='http://www.images.com/image2.jpg',
    )

    resource_client = LiveStyledResourceClient(TEST_API_DOMAIN, 'bar')

    with pytest.raises(ValueError, match='Cannot create a Team with an ID'):
        resource_client.create_team(team)


def test_update_team(requests_mock):
    team = Team(
        id=1,
        external_id='testteam1',
        name='Test Team',
        short_name='TT',
        light_crest_url='http://www.images.com/image1.jpg',
        dark_crest_url='http://www.images.com/image2.jpg',
    )

    resource_client = LiveStyledResourceClient(TEST_API_DOMAIN, 'bar')
    mock_responses = (
        ('PATCH', 'https://' + TEST_API_DOMAIN + '/v4/teams/1', 'mock_responses/ls_api/updated_team.json', 200),
    )
    configure_mock_responses(requests_mock, mock_responses, FIXTURES_DIR, CONTENT_TYPE)

    updated_team = resource_client.update_team(team, {'name': 'New Name'})

    for request in requests_mock.request_history:
        if request.method.upper() == 'PATCH' and request.url == 'https://' + TEST_API_DOMAIN + '/v4/teams/1':
            actual_payload = request.json()

            assert actual_payload == {
                'name': 'New Name'
            }

    assert updated_team.name == 'New Name'
    assert updated_team.id == 1
    assert updated_team.external_id == 'testteam1'
    assert updated_team.short_name == 'TT'
    assert updated_team.light_crest_url == 'http://www.images.com/image1.jpg'
    assert updated_team.dark_crest_url == 'http://www.images.com/image2.jpg'



class Season:
    def __init__(
            self,
            id: int,
            name: str or None,
            is_current: bool or None,
            external_id: str or None
    ):
        self.id = id
        self.name = name
        self.is_current = is_current
        self.external_id = external_id

    @classmethod
    def create_new(
            self,
            name: str,
            is_current: bool,
            external_id: str or None,
    ):
        self.id = None
        self.name = name
        self.is_current = is_current
        self.external_id = external_id

        return self

    @classmethod
    def placeholder(
            cls,
            id
    ):
        return cls(
            id=id,
            name=None,
            is_current=None,
            external_id=None
        )

    def diff(self, other):
        differences = {}
        fields = (
            'name', 'is_current', 'external_id'
        )
        for field in fields:
            if getattr(self, field) != getattr(other, field):
                differences[field] = getattr(self, field)
        return differences



from marshmallow import EXCLUDE, fields, Schema

from livestyled.models import DeviceToken
from livestyled.schemas.device import DeviceSchema
from livestyled.schemas.fields import RelatedResourceLinkField


class DeviceTokenSchema(Schema):
    class Meta:
        unknown = EXCLUDE
        api_type = 'device_tokens'
        url = 'v4/device_tokens'
        model = DeviceToken

    id = fields.Int()
    provider = fields.String()
    provider_token = fields.String(data_key='providerToken')
    payload = fields.Raw(allow_none=True)
    device_id = RelatedResourceLinkField(schema=DeviceSchema, data_key='device')



from datetime import datetime
from typing import Dict, List

from livestyled.models.device_consent import DeviceConsent
from livestyled.models.device_push_consent import DevicePushConsent


class Device:
    def __init__(
            self,
            id: int,
            token: str,
            consent: Dict,
            push_consents: List[Dict],
            type: str,
            status: str,
            app_version: str,
            os_version: str,
            model: str,
            manufacturer: str,
            bluetooth_on: bool,
            wifi_connected: bool,
            updated_at: datetime,
            created_at: datetime,
    ):
        self._id = id
        self.token = token
        if consent:
            self.consent = DeviceConsent(**consent)
        else:
            self.consent = None
        if push_consents:
            self.push_consents = [DevicePushConsent(**dpc) for dpc in push_consents]
        else:
            self.push_consents = []
        self.type = type
        self.status = status
        self.app_version = app_version
        self.os_version = os_version
        self.model = model
        self.manufacturer = manufacturer
        self.bluetooth_on = bluetooth_on
        self.wifi_connected = wifi_connected
        self.created_at = created_at
        self.updated_at = updated_at

    @property
    def id(self):
        return self._id

    @classmethod
    def placeholder(
            cls,
            id
    ):
        return cls(
            id=id,
            token=None,
            consent=None,
            push_consents=None,
            type=None,
            status=None,
            app_version=None,
            os_version=None,
            model=None,
            manufacturer=None,
            bluetooth_on=None,
            wifi_connected=None,
            updated_at=None,
            created_at=None,
        )

    def __repr__(self):
        return '<Device(id={self.id!r})>'.format(self=self)



from marshmallow import EXCLUDE, fields, Schema

from livestyled.models.magic_field import MagicField
from livestyled.schemas.fields import RelatedResourceLinkField


class MagicFieldSchema(Schema):
    class Meta:
        unknown = EXCLUDE
        api_type = 'magic_fields'
        url = 'v4/magic_fields'
        model = MagicField

    id = fields.Int()
    key = fields.String()
    value = fields.String()
    user_id = RelatedResourceLinkField(schema='livestyled.schemas.user.UserSchema', required=False, missing=None, data_key='user')



from marshmallow import EXCLUDE, fields, Schema

from livestyled.models.device_preference import DevicePreference
from livestyled.schemas.device import DeviceSchema
from livestyled.schemas.event import EventSchema
from livestyled.schemas.fields import RelatedResourceLinkField
from livestyled.schemas.venue import VenueSchema


class DevicePreferenceSchema(Schema):
    class Meta:
        unknown = EXCLUDE
        model = DevicePreference
        api_type = 'device_preferences'
        url = 'v4/device_preferences'
        default_ordering = '-createdAt'

    id = fields.Int()
    created_at = fields.AwareDateTime(data_key='createdAt')
    venue_id = RelatedResourceLinkField(schema=VenueSchema, data_key='venue', allow_none=True)
    device_id = RelatedResourceLinkField(schema=DeviceSchema, data_key='device')
    event_id = RelatedResourceLinkField(schema=EventSchema, data_key='event', allow_none=True)



from marshmallow import EXCLUDE, fields, Schema

from livestyled.models.fixture import Fixture
from livestyled.schemas.competition import CompetitionSchema
from livestyled.schemas.fields import RelatedResourceLinkField
from livestyled.schemas.season import SeasonSchema
from livestyled.schemas.sport_venue import SportVenueSchema
from livestyled.schemas.team import TeamSchema


class FixtureSchema(Schema):
    class Meta:
        unknown = EXCLUDE
        api_type = 'fixtures'
        url = 'v4/fixtures'
        model = Fixture

    class ScoreSchema(Schema):
        goals = fields.Int(allow_none=True)
        penalties = fields.Int(allow_none=True)

    class FixtureUrlSchema(Schema):
        class Meta:
            unknown = EXCLUDE

        title = fields.String(allow_none=True)
        url = fields.String(allow_none=True)
        is_enabled = fields.Boolean()

    id = fields.Int(load_only=True)
    start_at = fields.DateTime(data_key='startAt')
    is_fulltime = fields.Boolean(data_key='isFullTime')
    is_terminated = fields.Boolean(data_key='isTerminated')
    home_id = RelatedResourceLinkField(schema=TeamSchema, data_key='home')
    away_id = RelatedResourceLinkField(schema=TeamSchema, data_key='away')
    home_score = fields.Nested(ScoreSchema, data_key='homeScore')
    away_score = fields.Nested(ScoreSchema, data_key='awayScore')
    season_id = RelatedResourceLinkField(schema=SeasonSchema, data_key='season')
    competition_id = RelatedResourceLinkField(schema=CompetitionSchema, data_key='competition')
    venue_id = RelatedResourceLinkField(schema=SportVenueSchema, data_key='sportVenue')
    external_id = fields.String(missing=None, data_key='externalId')
    status = fields.String()
    url = fields.Nested(FixtureUrlSchema, allow_none=True, missing=None)
    allow_overwrite = fields.Boolean(data_key='allowOverwrite')



import json
import os

from livestyled.models.ticket_auth import TicketAuth
from livestyled.models.ticket_integration import TicketIntegration
from livestyled.models.user import UserEmail
from livestyled.resource_client import LiveStyledResourceClient
from livestyled.tests.utils import configure_mock_responses

FIXTURES_DIR = os.path.join(os.path.dirname(__file__), 'fixtures')
TEST_API_DOMAIN = 'test.livestyled.com'
CONTENT_TYPE = 'application/ld+json'


def test_get_ticket_auths(requests_mock):
    mock_responses = (
        ('GET', 'https://' + TEST_API_DOMAIN + '/v4/ticket_auths', 'mock_responses/ls_api/ticket_auth_list.json', 200),
    )
    configure_mock_responses(requests_mock, mock_responses, FIXTURES_DIR, CONTENT_TYPE)

    resource_client = LiveStyledResourceClient(TEST_API_DOMAIN, 'bar')
    ticket_auths = list(resource_client.get_ticket_auths())
    assert ticket_auths
    assert isinstance(ticket_auths[0], TicketAuth)
    assert ticket_auths[0].id == 1038
    assert isinstance(ticket_auths[0].user_email, UserEmail)
    assert ticket_auths[0].user_email.id == 330325
    assert ticket_auths[0].user_email.valid is True
    assert ticket_auths[0].user_email.email == 'amy.keane+verificatintest@livestyled.com'
    assert isinstance(ticket_auths[0].ticket_integration, TicketIntegration)
    assert ticket_auths[0].ticket_integration.id == 40
    assert ticket_auths[0].client_email == 'amy.keane+verificatintest@livestyled.com'
    assert ticket_auths[0].client_id == '129494829'


def test_get_ticket_auths_filtering(requests_mock):
    mock_responses = (
        ('GET', 'https://' + TEST_API_DOMAIN + '/v4/ticket_auths?userEmail.email=amy.keane%2Bverificatintest%40livestyled.com', 'mock_responses/ls_api/ticket_auth_list_filtered.json', 200),
    )
    configure_mock_responses(requests_mock, mock_responses, FIXTURES_DIR, CONTENT_TYPE)

    resource_client = LiveStyledResourceClient(TEST_API_DOMAIN, 'bar')
    ticket_auths = list(resource_client.get_ticket_auths(user_email_address='amy.keane+verificatintest@livestyled.com'))
    assert ticket_auths
    assert isinstance(ticket_auths[0], TicketAuth)
    assert ticket_auths[0].id == 1038
    assert isinstance(ticket_auths[0].user_email, UserEmail)
    assert ticket_auths[0].user_email.id == 330325
    assert ticket_auths[0].user_email.valid is True
    assert ticket_auths[0].user_email.email == 'amy.keane+verificatintest@livestyled.com'
    assert isinstance(ticket_auths[0].ticket_integration, TicketIntegration)
    assert ticket_auths[0].ticket_integration.id == 40
    assert ticket_auths[0].client_email == 'amy.keane+verificatintest@livestyled.com'
    assert ticket_auths[0].client_id == '129494829'


def test_update_ticket_auth(requests_mock):
    mock_responses = (
        ('PATCH', 'https://' + TEST_API_DOMAIN + '/v4/ticket_auths/1038', 'mock_responses/ls_api/ticket_auth_updated.json', 200),
    )
    configure_mock_responses(requests_mock, mock_responses, FIXTURES_DIR, CONTENT_TYPE)

    resource_client = LiveStyledResourceClient(TEST_API_DOMAIN, 'bar')
    resource_client.update_ticket_auth(
        TicketAuth.placeholder(1038),
        {
            'client_id': '99999'
        }
    )
    assert requests_mock.request_history[0].method == 'PATCH'
    assert json.loads(requests_mock.request_history[0].body) == {
        'clientId': '99999'
    }



import logging
from typing import Dict, Generator, List, Type

from marshmallow import Schema, ValidationError
from requests.exceptions import HTTPError

from livestyled.client import LiveStyledAPIClient
from livestyled.models import (
    Booking,
    Cohort,
    Competition,
    Device,
    DevicePreference,
    DeviceToken,
    Event,
    Fixture,
    LeagueTable,
    LeagueTableGroup,
    MagicField,
    News,
    Order,
    Product,
    ProductCategory,
    PushBroadcast,
    PushConsent,
    Season,
    SportVenue,
    Team,
    Ticket,
    TicketAuth,
    TicketIntegration,
    User,
    UserEmail,
    UserInfo,
    UserSSO,
    Venue,
)
from livestyled.schemas import (
    BookingSchema,
    CohortSchema,
    CompetitionSchema,
    DevicePreferenceSchema,
    DeviceSchema,
    DeviceTokenSchema,
    EventSchema,
    FixtureSchema,
    LeagueTableGroupSchema,
    LeagueTableSchema,
    MagicFieldSchema,
    NewsSchema,
    OrderSchema,
    ProductCategorySchema,
    ProductSchema,
    PushBroadcastSchema,
    PushConsentSchema,
    SeasonSchema,
    SportVenueSchema,
    TeamSchema,
    TicketAuthSchema,
    TicketIntegrationSchema,
    TicketSchema,
    UserCreateSchema,
    UserInfoSchema,
    UserSchema,
    UserSSOSchema,
    VenueSchema,
)

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

ASCENDING = 'asc'
DESCENDING = 'desc'


class LiveStyledResourceClient(LiveStyledAPIClient):
    def __init__(
            self,
            api_domain: str,
            api_key: str
    ):
        super(LiveStyledResourceClient, self).__init__(api_domain, api_key)

    def _get_resource_list(
            self,
            resource_schema: Type[Schema],
            external_id: str or None = None,
            filters: Dict or None = None,
            order_by: Dict or None = None,
    ):
        filter_params = {}
        if filters:
            filter_params = filters
        if external_id:
            filter_params['externalId'] = external_id

        if order_by:
            field = list(order_by.keys())[0]
            direction = list(order_by.values())[0]
            filter_params['order[{}]'.format(field)] = direction
            # TODO validate order parameters
        else:
            try:
                default_ordering = resource_schema.Meta.default_ordering
            except AttributeError:
                pass
            else:
                if default_ordering.startswith('-'):
                    filter_params['order[{}]'.format(default_ordering.lstrip('-'))] = 'desc'
                else:
                    filter_params['order[{}]'.format(default_ordering)] = 'asc'

        resources = self._get_resources(
            resource_schema,
            params=filter_params
        )
        for resource in resources:
            yield resource_schema.Meta.model(**resource)

    def _get_resource_by_id(
            self,
            schema: Type[Schema],
            id: int
    ):
        return schema.Meta.model(**self._get_resource(id, schema))

    def _create_resource(
            self,
            schema: Type[Schema],
            model_instance
    ):
        if getattr(model_instance, 'id', False) and model_instance.id is not None:
            raise ValueError('Cannot create a {} with an ID'.format(schema.Meta.model.__name__))
        payload = schema().dump(model_instance)
        for key, value in list(payload.items()):
            if value is None:
                payload.pop(key)
        new_instance = self._api_post(
            '{}'.format(schema.Meta.url),
            payload
        )
        try:
            return schema.Meta.model(**schema().load(new_instance))
        except ValidationError as err:
            logger.error(new_instance)
            logger.error(err.messages)
            raise

    def _update_resource(
            self,
            schema: Type[Schema],
            resource_id: int,
            attributes: Dict,
    ):
        attributes_to_update = list(attributes.keys())
        update_payload = schema(only=attributes_to_update).dump(attributes)
        updated_resource = self._api_patch(
            '{}/{}'.format(schema.Meta.url, resource_id),
            update_payload
        )
        return schema.Meta.model(**schema().load(updated_resource))

    def _delete_resource(
            self,
            schema: Type[Schema],
            resource,
    ):
        return self._api_delete(schema.Meta.url, resource.id)

    def get_teams(
            self,
            external_id: str or None = None,
    ) -> Generator[Team, None, None]:
        return self._get_resource_list(TeamSchema, external_id)

    # ---- TEAMS

    def get_team(
            self,
            id: int
    ) -> Team:
        return self._get_resource_by_id(TeamSchema, id)

    def create_team(
            self,
            team: Team
    ) -> Team:
        return self._create_resource(TeamSchema, team)

    def delete_team(
            self,
            team: Team
    ):
        self._delete_resource(TeamSchema, team)

    def update_team(
            self,
            team: Team,
            attributes: Dict
    ) -> Team:
        return self._update_resource(TeamSchema, team.id, attributes)

    # ---- FIXTURES

    def get_fixtures(
            self,
            external_id: str or None = None,
    ) -> Generator[Fixture, None, None]:
        return self._get_resource_list(FixtureSchema, external_id)

    def get_fixture(
            self,
            id: int
    ) -> Fixture:
        return self._get_resource_by_id(FixtureSchema, id)

    def create_fixture(
            self,
            fixture: Fixture
    ) -> Fixture:
        return self._create_resource(FixtureSchema, fixture)

    def update_fixture(
            self,
            fixture: Fixture,
            attributes: Dict
    ) -> Fixture:
        return self._update_resource(FixtureSchema, fixture.id, attributes)

    # ---- COMPETITIONS

    def get_competitions(
            self,
            external_id: str or None = None,
    ) -> Generator[Competition, None, None]:
        return self._get_resource_list(CompetitionSchema, external_id)

    def get_competition(
            self,
            id: int
    ) -> Competition:
        return self._get_resource_by_id(CompetitionSchema, id)

    def create_competition(
            self,
            competition: Competition
    ) -> Competition:
        return self._create_resource(CompetitionSchema, competition)

    def update_competition(
            self,
            competition: Competition,
            attributes: Dict
    ) -> Competition:
        return self._update_resource(CompetitionSchema, competition.id, attributes)

    # ---- SEASONS

    def get_seasons(
            self,
            external_id: str or None = None,
    ) -> Generator[Season, None, None]:
        return self._get_resource_list(SeasonSchema, external_id)

    def get_season(
            self,
            id: int
    ) -> Season:
        return self._get_resource_by_id(SeasonSchema, id)

    def create_season(
            self,
            season: Season
    ) -> Season:
        return self._create_resource(SeasonSchema, season)

    def update_season(
            self,
            season: Season,
            attributes: Dict
    ) -> Season:
        return self._update_resource(SeasonSchema, season.id, attributes)

    # ---- SPORTS VENUES

    def get_sport_venues(
            self,
            external_id: str or None = None,
    ) -> Generator[SportVenue, None, None]:
        return self._get_resource_list(SportVenueSchema, external_id)

    def get_sport_venue(
            self,
            id: int
    ) -> SportVenue:
        return self._get_resource_by_id(SportVenueSchema, id)

    def create_sport_venue(
            self,
            sport_venue: SportVenue
    ) -> SportVenue:
        return self._create_resource(SportVenueSchema, sport_venue)

    def update_sport_venue(
            self,
            sport_venue: SportVenue,
            attributes: Dict
    ) -> SportVenue:
        return self._update_resource(SportVenueSchema, sport_venue.id, attributes)

    # ---- LEAGUE TABLES

    def get_league_tables(
            self,
            external_id: str or None = None,
    ) -> Generator[LeagueTable, None, None]:
        return self._get_resource_list(LeagueTableSchema, external_id)

    def get_league_table(
            self,
            id: int
    ) -> LeagueTable:
        return self._get_resource_by_id(LeagueTableSchema, id)

    def create_league_table(
            self,
            league_table: LeagueTable
    ) -> LeagueTable:
        return self._create_resource(LeagueTableSchema, league_table)

    def update_league_table(
            self,
            league_table: LeagueTable,
            attributes: Dict
    ) -> LeagueTable:
        return self._update_resource(LeagueTableSchema, league_table.id, attributes)

    # ---- LEAGUE TABLE GROUPS

    def get_league_table_groups(
            self,
    ) -> Generator[LeagueTableGroup, None, None]:
        return self._get_resource_list(LeagueTableGroupSchema)

    def get_league_table_group(
            self,
            id: int
    ) -> LeagueTableGroup:
        return self._get_resource_by_id(LeagueTableGroupSchema, id)

    def create_league_table_group(
            self,
            league_table_group: LeagueTableGroup
    ) -> LeagueTableGroup:
        return self._create_resource(LeagueTableGroupSchema, league_table_group)

    def update_league_table_group(
            self,
            league_table: LeagueTableGroup,
            attributes: Dict
    ) -> LeagueTableGroup:
        return self._update_resource(LeagueTableGroupSchema, league_table.id, attributes)

    # ---- NEWS

    def get_news_articles(
            self,
            external_id: str or None = None,
    ) -> Generator[News, None, None]:
        return self._get_resource_list(NewsSchema, external_id)

    def get_news_article(
            self,
            id: int
    ) -> News:
        return self._get_resource_by_id(NewsSchema, id)

    def create_news_article(
            self,
            news: News
    ) -> News:
        return self._create_resource(NewsSchema, news)

    def update_news(
            self,
            news: News,
            attributes: Dict
    ) -> NewsSchema:
        return self._update_resource(NewsSchema, news.id, attributes)

    def delete_news(
            self,
            news: News,
    ) -> NewsSchema:
        return self._delete_resource(NewsSchema, news)

    # ---- USER

    def get_users(
            self,
            email: str or None = None,
    ) -> Generator[User, None, None]:
        if email:
            return self._get_resource_list(UserSchema, filters={'email': email})
        else:
            return self._get_resource_list(UserSchema)

    def get_user(
            self,
            id: int
    ) -> User:
        return self._get_resource_by_id(UserSchema, id)

    def create_user(
            self,
            user: User
    ) -> User:
        if user.id is not None:
            raise ValueError('Cannot create a User with an ID')
        payload = UserCreateSchema().dump(user)
        for key, value in list(payload.items()):
            if value is None:
                payload.pop(key)

        user_create_response = self._api_post(
            UserCreateSchema.Meta.create_url,
            payload
        )
        return User(**UserSchema().load(user_create_response))

    def update_user(
            self,
            user: User,
            attributes: Dict
    ) -> User:
        return self._update_resource(UserSchema, user.id, attributes)

    def update_user_info(
            self,
            user_info: UserInfo,
            attributes: Dict
    ) -> UserInfo:
        return self._update_resource(UserInfoSchema, user_info.id, attributes)

    def authorise_user(
            self,
            user: User,
            password: str
    ) -> Dict:
        if user.id is None:
            raise ValueError('Need a user ID to authorise a user')
        if not password:
            raise ValueError('Need a password to authorise a user')
        user_auth_response = self._api_post(
            UserSchema.Meta.authorise_url.format(user.id),
            {
                'password': password
            }
        )
        return user_auth_response

    # ---- USER SSO

    def get_user_ssos(
            self,
            sub: str or None = None,
    ) -> Generator[UserSSO, None, None]:
        if sub:
            return self._get_resource_list(UserSSOSchema, filters={'sub': sub})
        else:
            return self._get_resource_list(UserSSOSchema)

    def get_user_sso(
            self,
            id: int
    ) -> UserSSO:
        return self._get_resource_by_id(UserSSOSchema, id)

    def create_user_sso(
            self,
            user_sso: UserSSO
    ) -> UserSSO:
        return self._create_resource(UserSSOSchema, user_sso)

    def update_user_sso(
            self,
            user_sso: UserSSO,
            attributes: Dict
    ) -> UserSSO:
        return self._update_resource(UserSSOSchema, user_sso.id, attributes)

    def get_sso_for_user(
            self,
            user_id,
    ) -> UserSSO or None:
        try:
            user_sso_data = self._api_get('v4/users/{}/user_s_s_o'.format(user_id))
        except HTTPError as http_error:
            if http_error.response.status_code == 404:
                return None
            else:
                raise
        user_sso = UserSSO(**UserSSOSchema().load(user_sso_data))
        return user_sso

    # ---- PUSH BROADCASTS

    def get_push_broadcasts(
            self,
            order_by=None
    ) -> Generator[PushBroadcast, None, None]:
        return self._get_resource_list(PushBroadcastSchema, order_by=order_by)

    def get_push_broadcast(
            self,
            id: int
    ) -> PushBroadcast:
        return self._get_resource_by_id(PushBroadcastSchema, id)

    def create_push_broadcast(
            self,
            push_broadcast: PushBroadcast
    ) -> PushBroadcast:
        if push_broadcast.id is not None:
            raise ValueError('Cannot create a PushBroadcast with an ID')
        payload = PushBroadcastSchema().dump(push_broadcast)
        for key, value in list(payload.items()):
            if value is None:
                payload.pop(key)
        new_push_broadcast = self._api_post(
            PushBroadcastSchema.Meta.publish_url,
            payload
        )
        return PushBroadcast(**PushBroadcastSchema().load(new_push_broadcast))

    def update_push_broadcast(
            self,
            push_broadcast: PushBroadcast,
            attributes: Dict
    ) -> PushBroadcast:
        return self._update_resource(PushBroadcastSchema, push_broadcast.id, attributes)

    # ---- PUSH CONSENTS

    def get_push_consents(
            self,
    ) -> Generator[PushConsent, None, None]:
        return self._get_resource_list(PushConsentSchema)

    def get_push_consent(
            self,
            id: int
    ) -> PushConsent:
        return self._get_resource_by_id(PushConsentSchema, id)

    # ---- TICKETS

    def get_tickets(
            self,
            external_ticket_id: str or None = None,
            user: User or str or int or None = None,
    ) -> Generator[Ticket, None, None]:
        filters = {}
        if external_ticket_id:
            filters['externalTicketId'] = external_ticket_id
        if user:
            if isinstance(user, User):
                filters['user'] = '/v4/users/{}'.format(user.id)
            else:
                filters['user'] = '/v4/users/{}'.format(user)
        if filters:
            return self._get_resource_list(TicketSchema, filters=filters)
        else:
            return self._get_resource_list(TicketSchema)

    def get_ticket(
            self,
            id: int
    ) -> Ticket:
        return self._get_resource_by_id(TicketSchema, id)

    def create_ticket(
            self,
            ticket: Ticket
    ) -> Ticket:
        return self._create_resource(TicketSchema, ticket)

    def update_ticket(
            self,
            ticket: Ticket,
            attributes: Dict
    ) -> Ticket:
        return self._update_resource(TicketSchema, ticket.id, attributes)

    # ---- COHORTS

    def get_cohort(
            self,
            id: int
    ) -> Cohort:
        return self._get_resource_by_id(CohortSchema, id)

    def get_cohorts(
            self,
            external_id: str or None = None,
    ) -> Generator[Cohort, None, None]:
        if external_id:
            return self._get_resource_list(CohortSchema, filters={'externalId': external_id})
        else:
            return self._get_resource_list(CohortSchema)

    def create_cohort(
            self,
            cohort: Cohort
    ) -> Cohort:
        return self._create_resource(CohortSchema, cohort)

    def add_user_to_cohorts(
            self,
            user_id: int,
            cohorts: List[Cohort]
    ) -> bool:
        if not cohorts:
            return True
        payload = {
            'cohortExternalIds': [cohort.external_id for cohort in cohorts]
        }
        self._api_post(
            CohortSchema.Meta.bulk_user_attach_url.format(user_id),
            payload,
            content_type_override='application/json'
        )
        return True

    def remove_user_from_cohorts(
            self,
            user_id: int,
            cohorts: List[Cohort]
    ) -> bool:
        if not cohorts:
            return True
        payload = {
            'cohortExternalIds': [cohort.external_id for cohort in cohorts]
        }
        self._api_post(
            CohortSchema.Meta.bulk_user_detach_url.format(user_id),
            payload,
            content_type_override='application/json'
        )
        return True

    # ---- MAGIC FIELDS

    def get_magic_field(
            self,
            magic_field_id: int,
    ) -> MagicField:
        return self._get_resource_by_id(MagicFieldSchema, magic_field_id)

    def create_magic_field(
            self,
            magic_field: MagicField
    ) -> MagicField:
        return self._create_resource(MagicFieldSchema, magic_field)

    def update_magic_field(
            self,
            magic_field: MagicField,
            attributes: Dict
    ) -> MagicField:
        return self._update_resource(MagicFieldSchema, magic_field.id, attributes)

    # ---- EVENTS

    def get_events(
            self,
            title: str or None = None,
            status: str or None = None,
            start_at_after: str or None = None,
            start_at_before: str or None = None,
            end_at_after: str or None = None,
            external_id: str or None = None
    ) -> Generator[Event, None, None]:
        filters = {}
        if title:
            filters['title'] = title
        if status:
            filters['status'] = status
        if start_at_after:
            filters['event_dates.start_at[gt]'] = start_at_after
        if start_at_before:
            filters['event_dates.start_at[lt]'] = start_at_before
        if end_at_after:
            filters['event_dates.end_at[gt]'] = end_at_after
        if external_id:
            filters['externalId'] = external_id
        if filters:
            return self._get_resource_list(EventSchema, filters=filters)
        else:
            return self._get_resource_list(EventSchema)

    # ---- BOOKINGS

    def get_bookings(
            self,
            device: str or None = None,
            event: str or None = None,
            action: str or None = None,
            type: str or None = None,
    ) -> Generator[Booking, None, None]:
        filters = {}
        if device:
            if isinstance(device, Device):
                filters['device'] = '/v4/devices/{}'.format(device.id)
            else:
                filters['device'] = device
        if event:
            if isinstance(event, Event):
                filters['event'] = '/v4/events/{}'.format(event.id)
            else:
                filters['event'] = event
        if action:
            filters['action'] = action
        if type:
            filters['type'] = type
        if filters:
            return self._get_resource_list(BookingSchema, filters=filters)
        else:
            return self._get_resource_list(BookingSchema)

    def create_booking(
            self,
            booking: Booking
    ) -> Booking:
        return self._create_resource(BookingSchema, booking)

    def delete_booking(
            self,
            booking: Booking
    ) -> None:
        return self._delete_resource(BookingSchema, booking)

    # ---- DEVICE PREFERENCES

    def get_device_preferences(
            self,
            device: str or None = None,
            event: str or None = None,
    ) -> Generator[DevicePreference, None, None]:
        if device and event:
            return self._get_resource_list(DevicePreferenceSchema,
                                           filters={'device': device, 'event': event})
        else:
            return self._get_resource_list(DevicePreferenceSchema)

    def create_device_preference(
            self,
            device_preference: DevicePreference
    ) -> DevicePreference:
        return self._create_resource(DevicePreferenceSchema, device_preference)

    # ---- DEVICES

    def get_device(
            self,
            id,
    ) -> Device:
        return self._get_resource_by_id(DeviceSchema, id)

    def get_devices(
            self
    ) -> Generator[Device, None, None]:
        return self._get_resource_list(DeviceSchema)

    # ---- DEVICE TOKENS

    def get_tokens_for_device(
            self,
            device: Device,
            provider: str or None = None
    ) -> Generator[DeviceToken, None, None]:
        filters = {}
        if provider:
            filters['provider'] = provider
        try:
            device_token_data = self._api_get(
                'v4/devices/{}/device_tokens'.format(device.id),
                params=filters
            )
        except HTTPError as http_error:
            if http_error.response.status_code == 404:
                yield
            else:
                raise
        for device_token in device_token_data['hydra:member']:
            token = DeviceTokenSchema().load(device_token)
            yield DeviceTokenSchema.Meta.model(**token)

    def create_device_token(
            self,
            device_token: DeviceToken
    ) -> DeviceToken:
        return self._create_resource(DeviceTokenSchema, device_token)

    def update_device_token(
            self,
            device_token: DeviceToken,
            attributes: Dict
    ) -> DeviceToken:
        return self._update_resource(DeviceTokenSchema, device_token.id, attributes)

    # ---- ORDERS

    def get_order(
            self,
            id
    ) -> Order:
        return self._get_resource_by_id(OrderSchema, id)

    def get_orders(
            self,
    ) -> Generator[Order, None, None]:
        return self._get_resource_list(OrderSchema)

    def update_order(
            self,
            order: Order,
            attributes: Dict
    ) -> Order:
        return self._update_resource(OrderSchema, order.id, attributes)

    # -- PRODUCTS

    def get_product(
            self,
            id
    ) -> Product:
        return self._get_resource_by_id(ProductSchema, id)

    def get_products(
            self,
    ) -> Generator[Product, None, None]:
        return self._get_resource_list(ProductSchema)

    # ---- PRODUCT CATEGORIES

    def get_product_category(
            self,
            id
    ) -> ProductCategory:
        return self._get_resource_by_id(ProductCategorySchema, id)

    def get_product_categories(
            self,
    ) -> Generator[ProductCategory, None, None]:
        return self._get_resource_list(ProductCategorySchema)

    # ---- TICKET INTEGRATIONS

    def get_ticket_integrations(
            self,
    ) -> Generator[TicketIntegration, None, None]:
        return self._get_resource_list(TicketIntegrationSchema)

    def get_ticket_integration(
            self,
            id
    ) -> TicketIntegration:
        return self._get_resource_by_id(TicketIntegrationSchema, id)

    def create_ticket_integration(
            self,
            ticket_integration: TicketIntegration
    ) -> TicketIntegration:
        return self._create_resource(TicketIntegrationSchema, ticket_integration)

    # ---- TICKET AUTHS

    def get_ticket_auths(
            self,
            user_email_address=None,
            ticket_integration=None,
    ) -> Generator[TicketAuth, None, None]:
        filters = {}

        if user_email_address:
            if isinstance(user_email_address, str):
                filters['userEmail.email'] = user_email_address
            elif isinstance(user_email_address, UserEmail):
                filters['userEmail.email'] = user_email_address.email
        if ticket_integration:
            if isinstance(ticket_integration, (str, int)):
                filters['ticketIntegration'] = ticket_integration
            elif isinstance(ticket_integration, TicketIntegration):
                filters['ticketIntegration'] = ticket_integration.id

        if filters:
            return self._get_resource_list(TicketAuthSchema, filters=filters)
        else:
            return self._get_resource_list(TicketAuthSchema)

    def get_ticket_auth(
            self,
            id
    ) -> TicketAuth:
        return self._get_resource_by_id(TicketAuthSchema, id)

    def update_ticket_auth(
            self,
            ticket_auth: TicketAuth,
            attributes: Dict
    ) -> TicketAuth:
        return self._update_resource(TicketAuthSchema, ticket_auth.id, attributes)

    # ---- VENUES

    def get_venues(
            self,
            external_id: str or None = None,
    ) -> Generator[Venue, None, None]:
        return self._get_resource_list(VenueSchema)

    def get_venue(
            self,
            id: int
    ) -> Venue:
        return self._get_resource_by_id(VenueSchema, id)



from marshmallow import class_registry, fields
from marshmallow.base import SchemaABC


class RelatedResourceLinkField(fields.Field):

    def __init__(
            self,
            schema=None,
            many=False,
            **kwargs
    ):
        self._schema_arg = schema
        self.many = many
        self.__schema = None
        super(RelatedResourceLinkField, self).__init__(**kwargs)

    @property
    def schema(self):
        if not self.__schema and self._schema_arg:
            if isinstance(self._schema_arg, SchemaABC):
                self.__schema = self._schema_arg
            elif isinstance(self._schema_arg, type) and issubclass(self._schema_arg, SchemaABC):
                self.__schema = self._schema_arg
            elif isinstance(self._schema_arg, str):
                if self._schema_arg == 'self':
                    self.__schema = self.parent.__class__
                else:
                    self.__schema = class_registry.get_class(self._schema_arg)
            else:
                raise ValueError('Nested fields must be passed a Schema, not {0}.'.format(self.nested.__class__))
        return self.__schema

    def _serialize(self, value, attr, obj, **kwargs):
        if value:
            return '/{}/{}'.format(self.schema.Meta.url, value)
        return None

    def _deserialize(self, value, attr, data, **kwargs):
        if self.many:
            return [int(v.split('/')[-1]) for v in value]
        elif isinstance(value, dict):
            return int(value['id'])
        return int(value.split('/')[-1])


class RelatedResourceField(fields.Field):

    def __init__(
            self,
            schema=None,
            many=False,
            **kwargs
    ):
        self._schema_arg = schema
        self.many = many
        self.__schema = None
        super(RelatedResourceField, self).__init__(**kwargs)

    @property
    def schema(self):
        if not self.__schema and self._schema_arg:
            if isinstance(self._schema_arg, SchemaABC):
                self.__schema = self._schema_arg
            elif isinstance(self._schema_arg, type) and issubclass(self._schema_arg, SchemaABC):
                self.__schema = self._schema_arg
            elif isinstance(self._schema_arg, str):
                if self._schema_arg == 'self':
                    self.__schema = self.parent.__class__
                else:
                    self.__schema = class_registry.get_class(self._schema_arg)
            else:
                raise ValueError('Nested fields must be passed a Schema, not {0}.'.format(self.nested.__class__))
        return self.__schema

    def _serialize(self, value, attr, obj, **kwargs):
        if value:
            if not isinstance(value, (str, int)):
                value = getattr(value, 'id')
            return '/{}/{}'.format(self.schema.Meta.url, value)
        return None

    def _deserialize(self, value, attr, data, **kwargs):
        if self.many:
            deserialized = []
            for v in value:
                if isinstance(v, str):
                    deserialized.append(int(v.split('/')[-1]))
                elif isinstance(v, dict):
                    deserialized.append(self.schema().load(v))
            return deserialized
        else:
            return self.schema().load(value)



from marshmallow import EXCLUDE, fields, Schema

from livestyled.models.order import (
    Order,
    OrderItem,
)
from livestyled.schemas.app import AppSchema
from livestyled.schemas.fields import RelatedResourceField
from livestyled.schemas.fulfilment_point import FulfilmentPointSchema
from livestyled.schemas.product import ProductSchema, ProductVariantSchema
from livestyled.schemas.user import UserSchema


class OrderItemSchema(Schema):
    class Meta:
        unknown = EXCLUDE
        model = OrderItem

    id = fields.Int()
    product = RelatedResourceField(schema=ProductSchema, missing=None)
    product_variant = RelatedResourceField(schema=ProductVariantSchema, data_key='productVariant')
    fulfilment_point = RelatedResourceField(schema=FulfilmentPointSchema, data_key='fulfilmentPoint')
    quantity = fields.Int()
    title = fields.String(missing=None)
    subtitle = fields.String(data_key='subtitle', missing=None)
    image_url = fields.String(data_key='imageUrl', missing=None)
    price = fields.Integer()
    total_price = fields.Integer(data_key='totalPrice')


class OrderSchema(Schema):
    class Meta:
        unknown = EXCLUDE
        api_type = 'orders'
        url = 'v4/orders'
        model = Order

    id = fields.Int()
    app = fields.Nested(AppSchema, missing=None)
    user = fields.Nested(UserSchema, missing=None)
    status = fields.String(missing=None)
    gross_amount = fields.Integer(data_key='grossAmount', missing=None)
    discount = fields.Integer(missing=None)
    net_amount = fields.Integer(data_key='netAmount', missing=None)
    order_amount = fields.Integer(data_key='orderAmount', missing=None)
    order_number = fields.Integer(data_key='orderNumber', missing=None)
    items = RelatedResourceField(schema=OrderItemSchema, many=True)

    updated_at = fields.AwareDateTime(data_key='updatedAt', allow_none=True, missing=None)
    created_at = fields.AwareDateTime(data_key='createdAt', allow_none=True, missing=None)



from marshmallow import EXCLUDE, fields, Schema

from livestyled.models.fulfilment_point import (
    FulfilmentPoint,
    FulfilmentPointCategory,
    FulfilmentPointCategoryTranslation,
    FulfilmentPointTranslation,
)
from livestyled.schemas.fields import RelatedResourceField, RelatedResourceLinkField
from livestyled.schemas.venue import VenueSchema


class FulfilmentPointTranslationSchema(Schema):
    class Meta:
        unknown = EXCLUDE
        model = FulfilmentPointTranslation

    id = fields.Int()
    language = fields.String()
    title = fields.String(missing=None)
    description = fields.String(missing=None)
    collection_note = fields.String(missing=None, data_key='collectionNote')


class FulfilmentPointCategoryTranslationSchema(Schema):
    class Meta:
        unknown = EXCLUDE
        model = FulfilmentPointCategoryTranslation

    id = fields.Int()
    language = fields.String()
    title = fields.String(missing=None)


class FulfilmentPointCategorySchema(Schema):
    class Meta:
        unknown = EXCLUDE
        model = FulfilmentPointCategory

    id = fields.Int()
    status = fields.String()
    translations = RelatedResourceField(schema=FulfilmentPointCategoryTranslationSchema, many=True)


class FulfilmentPointSchema(Schema):
    class Meta:
        unknown = EXCLUDE
        model = FulfilmentPoint

    id = fields.Int()
    status = fields.String()
    image_url = fields.String(data_key='imageUrl', missing=None)
    map_image_url = fields.String(data_key='mapImageUrl', missing=None)
    lat = fields.String(missing=None)
    long = fields.String(missing=None)
    type = fields.String(missing=None)
    position = fields.Int(missing=None)
    reference = fields.String(missing=None)
    translations = RelatedResourceField(schema=FulfilmentPointTranslationSchema, many=True)
    categories = RelatedResourceField(schema=FulfilmentPointCategorySchema, many=True)
    venue = RelatedResourceLinkField(schema=VenueSchema, required=False, missing=None)



from datetime import datetime, timedelta, timezone
import json
import os

from livestyled.models.ticket import Ticket
from livestyled.models.ticket_integration import TicketIntegration
from livestyled.models.user import User
from livestyled.resource_client import LiveStyledResourceClient
from livestyled.tests.utils import configure_mock_responses

FIXTURES_DIR = os.path.join(os.path.dirname(__file__), 'fixtures')
TEST_API_DOMAIN = 'test.livestyled.com'
CONTENT_TYPE = 'application/ld+json'


def test_get_ticket(requests_mock):
    mock_responses = (
        ('GET', 'https://' + TEST_API_DOMAIN + '/v4/tickets/44', 'mock_responses/ls_api/example_ticket.json', 200),
    )
    configure_mock_responses(requests_mock, mock_responses, FIXTURES_DIR, CONTENT_TYPE)

    resource_client = LiveStyledResourceClient(TEST_API_DOMAIN, 'bar')
    ticket = resource_client.get_ticket(44)
    assert ticket
    assert isinstance(ticket, Ticket)
    assert ticket.id == 44
    assert ticket.external_ticket_id == '99999999'
    assert ticket.seat == '1'
    assert ticket.qr_code_url == ''
    assert ticket.session_date == datetime(2020, 12, 24, 0, 0, tzinfo=timezone(timedelta(0), '+0000'))
    assert ticket.barcode is None
    assert ticket.can_share is True
    assert ticket.client_email == 'test@livestyled.com'
    assert ticket.client_name is None
    assert ticket.created_at == datetime(2019, 5, 16, 15, 31, 45, tzinfo=timezone(timedelta(0), '+0000'))
    assert ticket.entrance is None
    assert ticket.external_customer_ref is None
    assert ticket.external_event_id is None
    assert ticket.parent_ticket is None
    assert ticket.premium is False
    assert ticket.price == 199
    assert ticket.price_code is None
    assert ticket.redeemer_email is None
    assert ticket.redeemer_id is None
    assert ticket.redeemer is None
    assert ticket.row == 'U'
    assert ticket.section == '128'
    assert ticket.sector_name is None
    assert ticket.share_code is None
    assert ticket.share_link is None
    assert ticket.shared_at is None
    assert ticket.sharer_email is None
    assert ticket.updated_at == datetime(2019, 5, 24, 12, 41, 22, tzinfo=timezone(timedelta(0), '+0000'))
    assert ticket.user_id is None
    assert ticket.user is None
    assert ticket.venue_name is None
    assert ticket.venue_room is None
    assert ticket.redeemed_at is None


def test_get_ticket_shared(requests_mock):
    mock_responses = (
        ('GET', 'https://' + TEST_API_DOMAIN + '/v4/tickets/44', 'mock_responses/ls_api/example_ticket_shared.json', 200),
    )
    configure_mock_responses(requests_mock, mock_responses, FIXTURES_DIR, CONTENT_TYPE)

    resource_client = LiveStyledResourceClient(TEST_API_DOMAIN, 'bar')
    ticket = resource_client.get_ticket(44)
    assert ticket
    assert isinstance(ticket, Ticket)
    assert ticket.id == 44
    assert ticket.external_ticket_id == '99999999'
    assert ticket.seat == '1'
    assert ticket.qr_code_url == ''
    assert ticket.session_date == datetime(2020, 12, 24, 0, 0, tzinfo=timezone(timedelta(0), '+0000'))
    assert ticket.barcode is None
    assert ticket.can_share is True
    assert ticket.client_email == 'test@livestyled.com'
    assert ticket.client_name is None
    assert ticket.created_at == datetime(2019, 5, 16, 15, 31, 45, tzinfo=timezone(timedelta(0), '+0000'))
    assert ticket.entrance is None
    assert ticket.external_customer_ref is None
    assert ticket.external_event_id is None
    assert ticket.parent_ticket is None
    assert ticket.premium is False
    assert ticket.price == 199
    assert ticket.price_code is None
    assert ticket.redeemer_email is None
    assert ticket.redeemer_id is None
    assert ticket.redeemer is None
    assert ticket.row == 'U'
    assert ticket.section == '128'
    assert ticket.sector_name is None
    assert ticket.share_code == 'ABCDEF12345'
    assert ticket.share_link is None
    assert ticket.shared_at == datetime(2019, 5, 24, 12, 41, 22, tzinfo=timezone(timedelta(0), '+0000'))
    assert ticket.sharer_email == 'test@livestyled.com'
    assert ticket.updated_at == datetime(2019, 5, 24, 12, 41, 22, tzinfo=timezone(timedelta(0), '+0000'))
    assert ticket.user_id is None
    assert ticket.user is None
    assert ticket.venue_name is None
    assert ticket.venue_room is None
    assert ticket.redeemed_at is None


def test_get_ticket_shared_redeemed(requests_mock):
    mock_responses = (
        ('GET', 'https://' + TEST_API_DOMAIN + '/v4/tickets/44', 'mock_responses/ls_api/example_ticket_shared_redeemed.json', 200),
    )
    configure_mock_responses(requests_mock, mock_responses, FIXTURES_DIR, CONTENT_TYPE)

    resource_client = LiveStyledResourceClient(TEST_API_DOMAIN, 'bar')
    ticket = resource_client.get_ticket(44)
    assert ticket
    assert isinstance(ticket, Ticket)
    assert ticket.id == 44
    assert ticket.external_ticket_id == '99999999'
    assert ticket.seat == '1'
    assert ticket.qr_code_url == ''
    assert ticket.session_date == datetime(2020, 12, 24, 0, 0, tzinfo=timezone(timedelta(0), '+0000'))
    assert ticket.barcode is None
    assert ticket.can_share is True
    assert ticket.client_email == 'test@livestyled.com'
    assert ticket.client_name is None
    assert ticket.created_at == datetime(2019, 5, 16, 15, 31, 45, tzinfo=timezone(timedelta(0), '+0000'))
    assert ticket.entrance is None
    assert ticket.external_customer_ref is None
    assert ticket.external_event_id is None
    assert ticket.parent_ticket is None
    assert ticket.premium is False
    assert ticket.price == 199
    assert ticket.price_code is None
    assert ticket.redeemer_email == 'someoneelse@livestyled.com'
    assert ticket.redeemer_id == 1234
    assert isinstance(ticket.redeemer, User)
    assert ticket.redeemer.id == 1234
    assert ticket.row == 'U'
    assert ticket.section == '128'
    assert ticket.sector_name is None
    assert ticket.share_code == 'ABCDEF12345'
    assert ticket.share_link is None
    assert ticket.shared_at == datetime(2019, 5, 24, 12, 41, 22, tzinfo=timezone(timedelta(0), '+0000'))
    assert ticket.sharer_email == 'test@livestyled.com'
    assert ticket.updated_at == datetime(2019, 5, 24, 13, 41, 22, tzinfo=timezone(timedelta(0), '+0000'))
    assert ticket.user_id is None
    assert ticket.user is None
    assert ticket.venue_name is None
    assert ticket.venue_room is None
    assert ticket.redeemed_at == datetime(2019, 5, 24, 13, 41, 22, tzinfo=timezone(timedelta(0), '+0000'))


def test_create_ticket(requests_mock):

    mock_responses = (
        ('POST', 'https://' + TEST_API_DOMAIN + '/v4/tickets', 'mock_responses/ls_api/create_ticket.json', 200),
    )
    configure_mock_responses(requests_mock, mock_responses, FIXTURES_DIR, CONTENT_TYPE)

    ticket_integration = TicketIntegration(
        label='TestTicketIntegration',
        id=99,
        name='TestTickets',
        module=None,
        endpoint_url=None,
        login_request=None,
        adapter=None,
        auth_required=None,
        default=None,
        can_share=None,
        config_payload=None
    )

    ticket = Ticket.create_new(
        ticket_integration=ticket_integration,
        user=User.placeholder(id='1234')
    )

    resource_client = LiveStyledResourceClient(TEST_API_DOMAIN, 'bar')

    created_ticket = resource_client.create_ticket(ticket)
    assert created_ticket.ticket_integration.id == 99

    create_request = requests_mock.request_history[0]
    assert create_request.method == 'POST'
    assert json.loads(create_request.body) == {
        'canShare': False,
        'premium': False,
        'price': 0,
        'user': '/v4/users/1234',
        'ticketIntegration': '/v4/ticket_integrations/99'
    }



from marshmallow import EXCLUDE, fields, Schema

from livestyled.models.sport_venue import SportVenue


class SportVenueSchema(Schema):
    class Meta:
        unknown = EXCLUDE
        api_type = 'sport_venues'
        url = 'v4/sport_venues'
        model = SportVenue

    id = fields.Int()
    name = fields.String()



from datetime import datetime, timedelta, timezone
import os

from livestyled.models import Product, ProductImage, ProductTranslation
from livestyled.resource_client import LiveStyledResourceClient
from livestyled.tests.utils import configure_mock_responses

FIXTURES_DIR = os.path.join(os.path.dirname(__file__), 'fixtures')
TEST_API_DOMAIN = 'test.livestyled.com'
CONTENT_TYPE = 'application/ld+json'


def test_get_products(requests_mock):
    mock_responses = (
        ('GET', 'https://' + TEST_API_DOMAIN + '/v4/products', 'mock_responses/ls_api/products.json', 200),
    )
    configure_mock_responses(requests_mock, mock_responses, FIXTURES_DIR, CONTENT_TYPE)

    resource_client = LiveStyledResourceClient(TEST_API_DOMAIN, 'bar')
    products = list(resource_client.get_products())
    assert len(products) == 30
    products.sort(key=lambda o: o.id)
    product = products[1]

    assert isinstance(product, Product)
    assert product.id == 2
    assert product.status == 'ACTIVE'
    assert product.reference == "Amy's product"
    assert len(product.images) == 1
    assert isinstance(product.images[0], ProductImage)
    assert product.images[0].id == 3
    assert product.images[0].position == 1
    assert product.images[0].created_at == datetime(2019, 7, 23, 11, 9, 57, tzinfo=timezone(timedelta(0), '+0000'))
    assert product.images[0].updated_at == datetime(2019, 11, 8, 11, 19, 31, tzinfo=timezone(timedelta(0), '+0000'))
    assert product.images[0].external_id is None
    assert product.images[0].image_url == 'https://apiv3.s3.eu-west-1.amazonaws.com/productimage/corgi%20bum.png'
    assert product.images[0].product.id == 2
    assert len(product.translations) == 1
    assert isinstance(product.translations[0], ProductTranslation)
    assert product.translations[0].id == 2
    assert product.translations[0].title == "Amy's product"
    assert product.translations[0].description == "A vegan burger from Beyond Meat with vegan smoked Gouda, Rubies in the Rubble Chipotle ‘Mayo’, mustard, red onion, pickles and lettuce.\r\n\r\nAfter a successful trial at our King's Cross restaurant, we're finally rolling out the Plant burger to every single Honest Burgers restaurant!\r\n\r\nThe Beyond Meat patty that goes into this burger is the result of over seven years of product development to look, cook and taste like beef. It also packs 20 grams of plant-based protein and has no GMOs, soy or gluten.\r\n\r\nA vegan burger from Beyond Meat with vegan smoked Gouda, Rubies in the Rubble Chipotle ‘Mayo’, mustard, red onion, pickles and lettuce.\r\n\r\nAfter a successful trial at our King's Cross restaurant, we're finally rolling out the Plant burger to every single Honest Burgers restaurant!\r\n\r\nThe Beyond Meat patty that goes into this burger is the result of over seven years of product development to look, cook and taste like beef. It also packs 20 grams of plant-based protein and has no GMOs, soy or gluten.\r\n\r\nA vegan burger from Beyond Meat with vegan smoked Gouda, Rubies in the Rubble Chipotle ‘Mayo’, mustard, red onion, pickles and lettuce.\r\n\r\nAfter a successful trial at our King's Cross restaurant, we're finally rolling out the Plant burger to every single Honest Burgers restaurant!\r\n\r\nThe Beyond Meat patty that goes into this burger is the result of over seven years of product development to look, cook and taste like beef. It also packs 20 grams of plant-based protein and has no GMOs, soy or gluten."
    assert product.translations[0].language == 'en'
    assert product.translations[0].created_at == datetime(2019, 7, 23, 11, 0, 2, tzinfo=timezone(timedelta(0), '+0000'))
    assert product.translations[0].updated_at == datetime(2019, 7, 25, 11, 30, 36, tzinfo=timezone(timedelta(0), '+0000'))



from livestyled.models.ticket_integration import TicketIntegration
from livestyled.models.user import UserEmail


class TicketAuth:
    def __init__(
            self,
            id,
            client_email,
            client_id,
            access_token,
            refresh_token,
            expire_at,
            user_email,
            ticket_integration
    ):
        self.id = id
        self.client_email = client_email
        self.client_id = client_id
        self.access_token = access_token
        self.refresh_token = refresh_token
        self.expire_at = expire_at
        if user_email:
            if isinstance(user_email, UserEmail):
                self.user_email = user_email
            elif isinstance(user_email, dict):
                self.user_email = UserEmail(**user_email)
            elif isinstance(user_email, (str, int)):
                self.user_email = UserEmail.placeholder(user_email)
        else:
            self.user_email = None
        if ticket_integration:
            if isinstance(ticket_integration, TicketIntegration):
                self.ticket_integration = ticket_integration
            elif isinstance(ticket_integration, dict):
                self.ticket_integration = TicketIntegration(**ticket_integration)
            elif isinstance(ticket_integration, (str, int)):
                self.ticket_integration = TicketIntegration.placeholder(ticket_integration)
        else:
            self.ticket_integration = None

    @classmethod
    def placeholder(
            cls,
            id
    ):
        return cls(
            id=id,
            client_id=None,
            client_email=None,
            access_token=None,
            refresh_token=None,
            expire_at=None,
            user_email=None,
            ticket_integration=None
        )

    @classmethod
    def create_new(
            cls,
            client_id,
            client_email,
            access_token,
            refresh_token,
            expire_at,
            user_email,
            ticket_integration
    ):
        return TicketAuth(
            id=None,
            client_id=client_id,
            client_email=client_email,
            access_token=access_token,
            refresh_token=refresh_token,
            expire_at=expire_at,
            user_email=user_email,
            ticket_integration=ticket_integration
        )

    def __repr__(self):
        return '<TicketAuth(id={self.id!r})>'.format(self=self)

    def diff(self, other):
        differences = {}
        fields = (
            'client_id', 'client_email', 'access_token', 'refresh_token', 'expire_at', 'user_email',
            'ticket_integration'
        )
        for field in fields:
            if getattr(self, field) != getattr(other, field):
                differences[field] = getattr(self, field)
        return differences



from marshmallow import EXCLUDE, fields, post_load, Schema

from livestyled.models.venue import Venue


class VenueSchema(Schema):
    class Meta:
        unknown = EXCLUDE
        api_type = 'venues'
        url = 'v4/venues'
        model = Venue

    id = fields.Int(missing=None)
    iri = fields.String(data_key='@id')
    name = fields.String()
    label = fields.String()
    status = fields.String()
    is_default = fields.Boolean(data_key='isDefault')
    description = fields.String(allow_none=True)
    image_url = fields.String(allow_none=True)
    map_image_url = fields.String(allow_none=True)
    geo_latitude = fields.String(allow_none=True)
    geo_longitude = fields.String(allow_none=True)
    geo_latitude_north_west = fields.String(allow_none=True)
    geo_longitude_north_west = fields.String(allow_none=True)
    geo_latitude_south_east = fields.String(allow_none=True)
    geo_longitude_south_west = fields.String(allow_none=True)
    city = fields.String(allow_none=True)
    external_id = fields.String(data_key='externalId')
    updated_at = fields.AwareDateTime(data_key='updatedAt', allow_none=True)
    created_at = fields.AwareDateTime(data_key='createdAt', allow_none=True)
    venue_icon_url = fields.String(data_key='venueIconUrl', allow_none=True)

    @post_load
    def fix_missing_id(self, data, **kwargs):
        if data['id'] is None:
            data['id'] = int(data['iri'].split('/')[-1])
            data.pop('iri', None)
        return data



from livestyled.models.event_date import EventDate
from livestyled.models.venue import Venue


class CoreEventCategory:
    def __init__(self, name):
        self.name = name


class Event:
    def __init__(
            self,
            id,
            status,
            title,
            description,
            promoted,
            updated_at,
            event_dates,
            created_at=None,
            translations=None,
            image_url=None,
            venues=None,
            core_event_category=None
    ):
        self._id = id
        self.status = status
        self.title = title
        self.description = description
        self.image_url = image_url
        self.promoted = promoted
        self.created_at = created_at
        self.updated_at = updated_at
        if event_dates:
            self.event_dates = [EventDate(**ed) for ed in event_dates]
        self.translations = translations
        if venues:
            self.venues = []
            for venue in venues:
                if isinstance(venue, dict):
                    self.venues.append(Venue(**venue))
                elif isinstance(venue, (str, int)):
                    self.venues.append(Venue.placeholder(id=venue))
                elif isinstance(venue, Venue):
                    self.venues.append(Venue)
        else:
            self.venues = []
        if core_event_category:
            if isinstance(core_event_category, dict):
                self.core_event_category = CoreEventCategory(name=core_event_category['name'])
            elif isinstance(core_event_category, CoreEventCategory):
                self.core_event_category = core_event_category
        else:
            self.core_event_category = CoreEventCategory(name='Other')

    @classmethod
    def placeholder(
            cls,
            id
    ):
        return cls(
            id=id,
            status=None,
            title=None,
            description=None,
            image_url=None,
            promoted=None,
            created_at=None,
            updated_at=None,
            event_dates=None,
            translations=None,
        )

    @property
    def id(self):
        return self._id

    def __repr__(self):
        return '<Event(id={self.id!r})>'.format(self=self)



from marshmallow import EXCLUDE, fields, Schema

from livestyled.models.cohort import Cohort


class CohortSchema(Schema):
    class Meta:
        unknown = EXCLUDE
        model = Cohort
        api_type = 'cohorts'
        url = 'v4/cohorts'
        bulk_user_attach_url = 'v4/users/{}/cohorts/bulk-attach'
        bulk_user_detach_url = 'v4/users/{}/cohorts/bulk-detach'

    title = fields.String()
    external_id = fields.String(data_key='externalId', required=False, missing=None)



class EventDate:
    def __init__(
            self,
            id,
            start_at,
            end_at,
            general_ticket_url=None,
    ):
        self._id = id
        self.start_at = start_at
        self.end_at = end_at
        self.general_ticket_url = general_ticket_url

    @property
    def id(self):
        return self._id

    def __repr__(self):
        return '<EventDate(id={self.id!r})>'.format(self=self)



from livestyled.models.device import Device
from livestyled.models.event import Event
from livestyled.models.venue import Venue


class DevicePreference:
    def __init__(
            self,
            id,
            created_at,
            device_id,
            venue_id,
            event_id,
    ):
        self._id = id
        self.created_at = created_at
        self._device = Device.placeholder(id=device_id)
        if venue_id:
            self._venue = Venue.placeholder(id=venue_id)
        else:
            self._venue = None
        if event_id:
            self._event = Event.placeholder(id=event_id)
        else:
            self._event = None

    @classmethod
    def create_new(
            cls,
            device: Device or str or int,
            venue: Venue or str or int or None,
            event: Event or str or int,
    ):
        device_pref = DevicePreference(
            id=None,
            device_id=None,
            venue_id=None,
            event_id=None,
            created_at=None,
        )
        if isinstance(event, (str, int)):
            event = Event.placeholder(event)
        device_pref._event = event
        if isinstance(device, (str, int)):
            device = Device.placeholder(id=device)
        device_pref._device = device
        if isinstance(venue, (str, int)):
            venue = Venue.placeholder(id=venue)
        device_pref._venue = venue
        return device_pref

    @property
    def id(self):
        return self._id

    @property
    def device_id(self):
        return self._device.id

    @property
    def device(self):
        return self._device

    @property
    def event_id(self):
        return self._event.id

    @property
    def event(self):
        return self._event

    @property
    def venue_id(self):
        return self._venue.id

    @property
    def venue(self):
        return self._venue

    def __repr__(self):
        return '<DevicePreference(id={self.id!r}, device={self.device!r}, venue={self.venue!r}, event={self.event!r})>'.format(self=self)

    def diff(self, other):
        differences = {}
        fields = (
            'device_id', 'venue_id', 'event_id',
        )
        for field in fields:
            if getattr(self, field) != getattr(other, field):
                differences[field] = getattr(self, field)
        return differences



from livestyled.models.team import Team


def test_team_difference():
    team_a = Team(id=1, name='Team A', short_name='TA', light_crest_url='light_url', external_id=1234, dark_crest_url='dark_url')
    team_b = Team(id=1, name='Team B', short_name='TA', light_crest_url='light_url', external_id=None, dark_crest_url='dark_url')

    difference = team_a.diff(team_b)
    assert difference == {
        'name': 'Team A',
        'external_id': 1234
    }



class Team:
    def __init__(
            self,
            id: int or None,
            name: str or None,
            short_name: str or None,
            light_crest_url: str or None,
            dark_crest_url: str or None,
            external_id: str or None
    ):
        self.id = id
        self.name = name
        self.short_name = short_name
        self.light_crest_url = light_crest_url
        self.dark_crest_url = dark_crest_url
        self.external_id = external_id

    @classmethod
    def placeholder(
            cls,
            id
    ):
        return cls(
            id=id,
            name=None,
            short_name=None,
            light_crest_url=None,
            dark_crest_url=None,
            external_id=None
        )

    @classmethod
    def create_new(
            cls,
            external_id: str,
            name: str,
            short_name: str,
            light_crest_url: str,
            dark_crest_url: str
    ):
        return Team(
            id=None,
            name=name,
            short_name=short_name,
            light_crest_url=light_crest_url,
            dark_crest_url=dark_crest_url,
            external_id=external_id
        )

    def diff(self, other):
        differences = {}
        fields = (
            'name', 'short_name', 'light_crest_url', 'external_id', 'dark_crest_url'
        )
        for field in fields:
            if getattr(self, field) != getattr(other, field):
                differences[field] = getattr(self, field)
        return differences



from livestyled.models.app import App
from livestyled.models.fulfilment_point import FulfilmentPoint
from livestyled.models.product import Product, ProductVariant
from livestyled.models.user import User


class OrderItem:
    def __init__(
            self,
            id,
            product_variant,
            fulfilment_point,
            quantity,
            title,
            subtitle,
            image_url,
            price,
            total_price,
            product,
    ):
        self.id = id
        self.quantity = quantity
        self.title = title
        self.subtitle = subtitle
        self.image_url = image_url
        self.price = price
        self.total_price = total_price

        if product_variant:
            if isinstance(product_variant, ProductVariant):
                self.product_variant = product_variant
            elif isinstance(product_variant, dict):
                self.product_variant = ProductVariant(**product_variant)
        else:
            self.product_variant = None

        if fulfilment_point:
            if isinstance(fulfilment_point, FulfilmentPoint):
                self.fulfilment_point = fulfilment_point
            elif isinstance(fulfilment_point, dict):
                self.fulfilment_point = FulfilmentPoint(**fulfilment_point)
        else:
            self.fulfilment_point = None

        if product:
            if isinstance(product, Product):
                self.product = product
            elif isinstance(product, dict):
                self.product = Product(**product)
        else:
            self.product = None


class Order:
    def __init__(
            self,
            id,
            user,
            status,
            gross_amount,
            discount,
            net_amount,
            order_amount,
            order_number,
            items,
            updated_at,
            created_at,
            app,
    ):
        self.id = id
        self.status = status
        self.gross_amount = gross_amount
        self.discount = discount
        self.net_amount = net_amount
        self.order_amount = order_amount
        self.updated_at = updated_at
        self.created_at = created_at
        self.order_number = order_number

        if user:
            if isinstance(user, User):
                self.user = user
            elif isinstance(user, dict):
                self.user = User(**user)
        else:
            self.user = None

        if items:
            self.items = []
            for item in items:
                if isinstance(item, OrderItem):
                    self.items.append(item)
                elif isinstance(item, dict):
                    self.items.append(OrderItem(**item))
        else:
            self.items = None

        if app:
            if isinstance(app, App):
                self.app = app
            elif isinstance(app, dict):
                self.app = App(**app)
        else:
            self.app = None



from marshmallow import EXCLUDE, fields, Schema

from livestyled.models.user import User, UserConsent, UserEmail, UserInfo, UserSSO
from livestyled.schemas.cohort import CohortSchema
from livestyled.schemas.device import DeviceSchema
from livestyled.schemas.fields import RelatedResourceField, RelatedResourceLinkField
from livestyled.schemas.magic_field import MagicFieldSchema


class UserCreateSchema(Schema):
    class Meta:
        api_type = 'users'
        url = 'v4/users'
        create_url = 'v4/users/register'
        model = User

    auth_type = fields.String(data_key='authType')
    first_name = fields.String(data_key='firstName')
    last_name = fields.String(data_key='lastName')
    email = fields.Email()
    password = fields.String()
    device_id = RelatedResourceLinkField(schema=DeviceSchema, data_key='device')


class UserInfoSchema(Schema):
    class Meta:
        unknown = EXCLUDE
        api_type = 'user_info'
        url = 'v4/user_infos'
        model = UserInfo

    id = fields.Int()
    first_name = fields.String(data_key='firstName', missing=None)
    last_name = fields.String(data_key='lastName', missing=None)
    phone = fields.String(data_key='phone', missing=None)
    dob = fields.AwareDateTime(missing=None)
    gender = fields.String(missing=None)


class UserConsentSchema(Schema):
    class Meta:
        unknown = EXCLUDE
        model = UserConsent

    id = fields.Int()
    marketing_consent = fields.Boolean(data_key='marketingConsent', allow_none=False)
    analysis_consent = fields.Boolean(data_key='analysisConsent', allow_none=False)


class UserEmailSchema(Schema):
    class Meta:
        unknown = EXCLUDE
        model = UserEmail

    valid = fields.Boolean()
    email = fields.String()
    id = fields.Int()


class UserSchema(Schema):
    class Meta:
        unknown = EXCLUDE
        api_type = 'users'
        url = 'v4/users'
        create_url = 'v4/users/register'
        authorise_url = 'v4/users/{}/authorise'
        magic_fields_url = 'v4/users/{}/magic_fields'
        model = User

    id = fields.Int()
    auth_type = fields.String(data_key='authType', missing=None)
    email = fields.Email()
    password = fields.String()
    device_id = RelatedResourceLinkField(schema=DeviceSchema, data_key='device')
    user_info = fields.Nested(UserInfoSchema, data_key='userInfo')
    cohorts = RelatedResourceLinkField(schema=CohortSchema, many=True)
    magic_fields = RelatedResourceField(schema=MagicFieldSchema, many=True, data_key='magicFields')
    devices = RelatedResourceField(schema=DeviceSchema, many=True, data_key='devices')
    user_emails = RelatedResourceField(schema=UserEmailSchema, data_key='userEmails', many=True)
    user_consent = RelatedResourceField(schema=UserConsentSchema, data_key='userConsent')
    token = fields.String(missing=None)


class UserSSOSchema(Schema):
    class Meta:
        unknown = EXCLUDE
        api_type = 'user_sso'
        url = 'v4/user_ssos'
        model = UserSSO

    id = fields.Int()
    access_token = fields.String(data_key='accessToken', missing=None, allow_none=True)
    refresh_token = fields.String(data_key='refreshToken', missing=None, allow_none=True)
    sub = fields.String()
    expires = fields.AwareDateTime(missing=None, allow_none=True)
    user_id = RelatedResourceLinkField(schema=UserSchema, data_key='user')



from datetime import datetime
from typing import Dict, List

from livestyled.models.cohort import Cohort
from livestyled.models.device import Device
from livestyled.models.magic_field import MagicField


class UserInfo:
    def __init__(
            self,
            id: int or None = None,
            first_name: str or None = None,
            last_name: str or None = None,
            dob: datetime or None = None,
            gender: str or None = None,
            phone: str or None = None,
    ):
        self.id = id
        self.first_name = first_name
        self.last_name = last_name
        self.dob = dob
        self.gender = gender
        self.phone = phone

    def __hash__(self):
        return hash((self.first_name, self.last_name, self.dob, self.gender, self.phone))

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return all(
            [
                self.first_name == other.first_name,
                self.last_name == other.last_name,
                self.dob == other.dob,
                self.gender == other.gender,
                self.phone == other.phone
            ]
        )


class UserConsent:
    def __init__(
            self,
            id: int or None = None,
            marketing_consent: bool = None,
            analysis_consent: bool = None
    ):
        self.id = id
        self.marketing_consent = marketing_consent
        self.analysis_consent = analysis_consent

    def __hash__(self):
        return hash((self.marketing_consent, self.analysis_consent))

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented

        return all(
            [
                self.marketing_consent == other.marketing_consent,
                self.analysis_consent == other.analysis_consent
            ]
        )


class UserEmail:
    def __init__(
            self,
            id: int,
            valid: bool,
            email: str
    ):
        self.id = id
        self.valid = valid
        self.email = email


class User:
    def __init__(
            self,
            id: int or None,
            email: str or None,
            auth_type: str or None,
            user_info: dict or None,
            device_id: str or None = None,
            password: str or None = None,
            cohorts: List[int] or None = None,
            magic_fields: List[Dict] or None = None,
            user_emails: List[Dict] or None = None,
            user_consent: Dict or None = None,
            devices: List[Dict] or List[str] or None = None,
            token=None,
    ):
        self.id = id
        self.email = email
        self.auth_type = auth_type
        self._device = Device.placeholder(id=device_id)
        if user_info:
            if isinstance(user_info, dict):
                user_info = UserInfo(**user_info)
            elif isinstance(user_info, UserInfo):
                user_info = user_info
        else:
            user_info = UserInfo(None, None, None, None)
        self.user_info = user_info
        self.first_name = user_info.first_name
        self.last_name = user_info.last_name
        self.password = password
        self.cohorts = []
        self.token = token
        if cohorts:
            self.cohorts = [Cohort.placeholder(int(c_id)) for c_id in cohorts]
        if magic_fields:
            self.magic_fields = [MagicField(**m) for m in magic_fields]
        else:
            self.magic_fields = []
        if user_emails:
            self.user_emails = [UserEmail(**u) for u in user_emails]
        else:
            self.user_emails = []
        if user_consent:
            if isinstance(user_consent, UserConsent):
                self.user_consent = user_consent
            elif isinstance(user_consent, dict):
                self.user_consent = UserConsent(**user_consent)
        else:
            self.user_consent = None
        if devices:
            self.devices = []
            for device in devices:
                if isinstance(device, (str, int)):
                    self.devices.append(Device.placeholder(id=device))
                elif isinstance(device, dict):
                    self.devices.append(Device(**device))
        else:
            self.devices = []

    @classmethod
    def create_new(
            cls,
            email: str,
            auth_type: str,
            first_name: str,
            last_name: str,
            device: str,
            password: str or None,
            dob: datetime or None = None,
            gender: str or None = None,
            marketing_consent: bool = False,
            analysis_consent: bool = False
    ):
        user = User(
            id=None,
            email=email,
            auth_type=auth_type,
            user_info=UserInfo(first_name=first_name, last_name=last_name, dob=dob, gender=gender),
            device_id=None,
            password=password,
            user_consent=UserConsent(marketing_consent=marketing_consent, analysis_consent=analysis_consent)
        )
        user._device = device
        return user

    @classmethod
    def placeholder(
            cls,
            id
    ):
        return cls(
            id=id,
            email=None,
            auth_type=None,
            user_info=UserInfo(),
            device_id=None,
            password=None,
            user_consent=UserConsent()
        )

    def diff(self, other):
        differences = {}
        fields = (
            'email', 'auth_type', 'user_info',
        )
        for field in fields:
            if getattr(self, field) != getattr(other, field):
                differences[field] = getattr(self, field)
        return differences

    @property
    def device_id(self):
        return self._device.id

    @property
    def device(self):
        return self._device


class UserSSO:
    def __init__(
            self,
            id,
            user_id,
            access_token,
            refresh_token,
            expires,
            sub
    ):
        self.id = id
        self.access_token = access_token
        self.refresh_token = refresh_token
        self.expires = expires
        self.sub = sub
        self._user = User.placeholder(user_id)

    @classmethod
    def create_new(
            cls,
            user,
            access_token,
            refresh_token,
            expires,
            sub
    ):
        user_sso = UserSSO(
            id=None,
            user_id=None,
            access_token=access_token,
            refresh_token=refresh_token,
            expires=expires,
            sub=sub
        )
        user_sso._user = user
        return user_sso

    def diff(self, other):
        differences = {}
        fields = (
            'access_token', 'refresh_token', 'expires', 'sub', 'user_id',
        )
        for field in fields:
            if getattr(self, field) != getattr(other, field):
                differences[field] = getattr(self, field)
        return differences

    @property
    def user_id(self):
        return self._user.id

    @property
    def user(self):
        return self._user



from marshmallow import EXCLUDE, fields, Schema

from livestyled.models import Device, DevicePushConsent
from livestyled.schemas.device_consent import DeviceConsentSchema
from livestyled.schemas.fields import RelatedResourceField, RelatedResourceLinkField
from livestyled.schemas.push_consent import PushConsentSchema


class DevicePushConsentSchema(Schema):
    class Meta:
        unknown = EXCLUDE
        model = DevicePushConsent

    id = fields.Int()
    consent = fields.Boolean()
    push_consent_id = RelatedResourceLinkField(schema=PushConsentSchema, data_key='pushConsent')


class DeviceSchema(Schema):
    class Meta:
        unknown = EXCLUDE
        api_type = 'devices'
        url = 'v4/devices'
        model = Device

    id = fields.Int()
    token = fields.String()
    consent = RelatedResourceField(schema=DeviceConsentSchema, allow_none=True, missing=None)
    push_consents = RelatedResourceField(schema=DevicePushConsentSchema, many=True, data_key='pushConsents', required=False, missing=None)
    type = fields.String()
    status = fields.String()
    app_version = fields.String(data_key='appVersion')
    os_version = fields.String(data_key='osVersion', allow_none=True, missing=None)
    model = fields.String(allow_none=True, missing=None)
    manufacturer = fields.String(allow_none=True, missing=None)
    bluetooth_on = fields.Boolean(data_key='bluetoothOn', allow_none=True, missing=None)
    wifi_connected = fields.Boolean(data_key='wifiConnected', allow_none=True, missing=None)
    updated_at = fields.AwareDateTime(data_key='updatedAt', allow_none=True, missing=None)
    created_at = fields.AwareDateTime(data_key='createdAt', allow_none=True, missing=None)



import os

from livestyled.models.device import Device
from livestyled.resource_client import LiveStyledResourceClient
from livestyled.tests.utils import configure_mock_responses

FIXTURES_DIR = os.path.join(os.path.dirname(__file__), 'fixtures')
TEST_API_DOMAIN = 'test.livestyled.com'
CONTENT_TYPE = 'application/ld+json'


def test_get_device_by_id(requests_mock):
    mock_responses = (
        ('GET', 'https://' + TEST_API_DOMAIN + '/v4/devices/1234', 'mock_responses/ls_api/device_1234.json', 200),
    )
    configure_mock_responses(requests_mock, mock_responses, FIXTURES_DIR, CONTENT_TYPE)

    resource_client = LiveStyledResourceClient(TEST_API_DOMAIN, 'bar')
    device = resource_client.get_device(1234)
    assert device
    assert isinstance(device, Device)
    assert device.id == 1234
    assert device.push_consents == []



from livestyled.models.device import Device
from livestyled.models.event import Event
from livestyled.models.user import User


class Booking:
    def __init__(
            self,
            id,
            type,
            device_id,
            event_id,
            created_at,
            updated_at,
            action,
            user_id=None,
    ):
        self._id = id
        self.type = type
        self._device = Device.placeholder(id=device_id)
        if user_id:
            self._user = User.placeholder(id=user_id)
        else:
            self._user = None
        self._event = Event.placeholder(id=event_id)
        self.created_at = created_at
        self.updated_at = updated_at
        self.action = action

    @classmethod
    def create_new(
            cls,
            type: str,
            device: Device or str or int,
            user: User or str or int or None,
            event: Event or str or int,
            action: str,
    ):
        booking = Booking(
            id=None,
            type=type,
            device_id=None,
            user_id=None,
            event_id=None,
            created_at=None,
            updated_at=None,
            action=action,
        )
        if isinstance(user, (str, int)):
            user = User.placeholder(id=user)
        booking._user = user
        if isinstance(device, (str, int)):
            device = Device.placeholder(id=device)
        booking._device = device
        if isinstance(event, (str, int)):
            event = Event.placeholder(id=event)
        booking._event = event
        return booking

    @property
    def id(self):
        return self._id

    @property
    def device_id(self):
        return self._device.id

    @property
    def device(self):
        return self._device

    @property
    def user_id(self):
        if self._user:
            return self._user.id
        else:
            return None

    @property
    def user(self):
        return self._user

    @property
    def event_id(self):
        return self._event.id

    @property
    def event(self):
        return self._event

    def __repr__(self):
        return '<Booking(id={self.id!r}, type={self.type!r}, action={self.action!r})>'.format(self=self)

    def diff(self, other):
        differences = {}
        fields = (
            'device_id', 'user_id', 'event_id', 'type', 'action'
        )
        for field in fields:
            if getattr(self, field) != getattr(other, field):
                differences[field] = getattr(self, field)
        return differences



import json
import os

from livestyled.models.order import Order
from livestyled.utils import create_resource_from_data

FIXTURES_DIR = os.path.join(os.path.dirname(__file__), 'fixtures')


def test_create_order_model():
    with open(os.path.join(FIXTURES_DIR, 'order.json'), 'r') as fixture_file:
        order_data = json.loads(fixture_file.read())

    order = create_resource_from_data(Order, order_data)

    assert isinstance(order, Order)



import os
import re

from pkg_resources import parse_requirements
from setuptools import find_packages, setup

del os.link

PACKAGE_NAME = 'livestyled'


def parse_version(package_name):
    """
    if environment variable PACKAGE_VERSION set use that as version number
    else open `VERSION_FILE` and find `__version__ = x.y.z`.
    thanks to http://stackoverflow.com/a/7071358
    """
    version = os.environ.get('PACKAGE_VERSION', None)
    if version:
        return version

    VERSION_RE = r"^__version__ = ['\"]([^'\"]*)['\"]"
    VERSION_FILE = '{}/__init__.py'.format(package_name)
    with open(VERSION_FILE, 'r') as f:
        file_contents = f.read()
    match = re.search(VERSION_RE, file_contents, re.M)
    if match:
        return match.group(1)
    else:
        raise RuntimeError('Unable to find version string in {}.'.format(VERSION_FILE))


requirements_fn = os.path.join(os.path.dirname(__file__), 'requirements.txt')

setup(
    name=PACKAGE_NAME,
    version=parse_version(PACKAGE_NAME),
    packages=find_packages(),
    include_package_data=True,
    author='LiveStyled',
    author_email='dev@livestyled.com',
    url='https://github.com/livestyled/python-sdk',
    install_requires=[str(req) for req in parse_requirements(open(requirements_fn).read())],
    entry_points="""
    [console_scripts]
    """
)
